# /********************************************************************************
# * Copyright (c) 2024 Contributors to the Eclipse Foundation
# *
# * See the NOTICE file(s) distributed with this work for additional
# * information regarding copyright ownership.
# *
# * This program and the accompanying materials are made available under the
# * terms of the Apache License 2.0 which is available at
# * http://www.apache.org/licenses/LICENSE-2.0
# *
# * SPDX-License-Identifier: Apache-2.0
# ********************************************************************************/

# This file contains a logger that could be useful for clients using kuksa-python-sdk

import logging
import sys
import os


class KuksaLogger:
    def init_logging(self):
        # Example
        #
        # Set log level to debug
        #   LOG_LEVEL=debug kuksa-client
        #
        # Set log level to INFO, but for dbcfeederlib.databrokerclientwrapper set it to DEBUG
        #   LOG_LEVEL=info,dbcfeederlib.databrokerclientwrapper=debug kuksa-client
        #
        #

        loglevels = self.parse_env_log(os.environ.get("LOG_LEVEL"))

        # set root loglevel etc
        self.init_root_logging(loglevels["root"])

        # set loglevels for other loggers
        for logger, level in loglevels.items():
            if logger != "root":
                logging.getLogger(logger).setLevel(level)

    def init_root_logging(self, loglevel):
        """Set up console logger"""
        # create console handler and set level to debug. This just means that it can show DEBUG messages.
        # What actually is shown is controlled by logging configuration
        console_logger = logging.StreamHandler()
        console_logger.setLevel(logging.DEBUG)

        # create formatter
        if sys.stdout.isatty():
            formatter = ColorFormatter()
        else:
            formatter = logging.Formatter(
                fmt="%(asctime)s %(levelname)s %(name)s: %(message)s"
            )

        # add formatter to console_logger
        console_logger.setFormatter(formatter)

        # add console_logger as a global handler
        root_logger = logging.getLogger()
        root_logger.setLevel(loglevel)
        root_logger.addHandler(console_logger)

    def parse_env_log(self, env_log, default=logging.INFO):
        def parse_level(specified_level, default=default):
            if isinstance(specified_level, str):
                if specified_level.lower() in [
                    "debug",
                    "info",
                    "warn",
                    "warning",
                    "error",
                    "critical",
                                               ]:
                    return specified_level.upper()
                raise ValueError(f"could not parse '{specified_level}' as a log level")
            return default

        parsed_loglevels = {}

        if env_log is not None:
            log_specs = env_log.split(",")
            for log_spec in log_specs:
                spec_parts = log_spec.split("=")
                if len(spec_parts) == 1:
                    # This is a root level spec
                    if "root" in parsed_loglevels:
                        raise ValueError("multiple root loglevels specified")
                    parsed_loglevels["root"] = parse_level(spec_parts[0])
                if len(spec_parts) == 2:
                    logger_name = spec_parts[0]
                    logger_level = spec_parts[1]
                    parsed_loglevels[logger_name] = parse_level(logger_level)

        if "root" not in parsed_loglevels:
            parsed_loglevels["root"] = default

        return parsed_loglevels


class ColorFormatter(logging.Formatter):
    """Color formatter that can be used for terminals"""
    FORMAT = "{time} {{loglevel}} {logger} {msg}".format(
        time="\x1b[2m%(asctime)s\x1b[0m",  # grey
        logger="\x1b[2m%(name)s:\x1b[0m",  # grey
        msg="%(message)s",
    )
    FORMATS = {
        logging.DEBUG: FORMAT.format(loglevel="\x1b[34mDEBUG\x1b[0m"),  # blue
        logging.INFO: FORMAT.format(loglevel="\x1b[32mINFO\x1b[0m"),  # green
        logging.WARNING: FORMAT.format(loglevel="\x1b[33mWARNING\x1b[0m"),  # yellow
        logging.ERROR: FORMAT.format(loglevel="\x1b[31mERROR\x1b[0m"),  # red
        logging.CRITICAL: FORMAT.format(loglevel="\x1b[31mCRITICAL\x1b[0m"),  # red
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from sdv.databroker.v1 import broker_pb2 as sdv_dot_databroker_dot_v1_dot_broker__pb2

GRPC_GENERATED_VERSION = '1.68.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in sdv/databroker/v1/broker_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class BrokerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetDatapoints = channel.unary_unary(
                '/sdv.databroker.v1.Broker/GetDatapoints',
                request_serializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.GetDatapointsRequest.SerializeToString,
                response_deserializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.GetDatapointsReply.FromString,
                _registered_method=True)
        self.SetDatapoints = channel.unary_unary(
                '/sdv.databroker.v1.Broker/SetDatapoints',
                request_serializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.SetDatapointsRequest.SerializeToString,
                response_deserializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.SetDatapointsReply.FromString,
                _registered_method=True)
        self.Subscribe = channel.unary_stream(
                '/sdv.databroker.v1.Broker/Subscribe',
                request_serializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.SubscribeRequest.SerializeToString,
                response_deserializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.SubscribeReply.FromString,
                _registered_method=True)
        self.GetMetadata = channel.unary_unary(
                '/sdv.databroker.v1.Broker/GetMetadata',
                request_serializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.GetMetadataRequest.SerializeToString,
                response_deserializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.GetMetadataReply.FromString,
                _registered_method=True)


class BrokerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetDatapoints(self, request, context):
        """Request a set of datapoints (values)

        Returns a list of requested data points.

        InvalidArgument is returned if the request is malformed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetDatapoints(self, request, context):
        """Set a datapoint (values)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Subscribe(self, request, context):
        """Subscribe to a set of data points or conditional expressions
        using the Data Broker Query Syntax (described in QUERY.md)

        Returns a stream of replies.

        InvalidArgument is returned if the request is malformed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMetadata(self, request, context):
        """Request the metadata of a set of datapoints

        Returns metadata of the requested data points that exist.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BrokerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetDatapoints': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDatapoints,
                    request_deserializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.GetDatapointsRequest.FromString,
                    response_serializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.GetDatapointsReply.SerializeToString,
            ),
            'SetDatapoints': grpc.unary_unary_rpc_method_handler(
                    servicer.SetDatapoints,
                    request_deserializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.SetDatapointsRequest.FromString,
                    response_serializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.SetDatapointsReply.SerializeToString,
            ),
            'Subscribe': grpc.unary_stream_rpc_method_handler(
                    servicer.Subscribe,
                    request_deserializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.SubscribeRequest.FromString,
                    response_serializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.SubscribeReply.SerializeToString,
            ),
            'GetMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMetadata,
                    request_deserializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.GetMetadataRequest.FromString,
                    response_serializer=sdv_dot_databroker_dot_v1_dot_broker__pb2.GetMetadataReply.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'sdv.databroker.v1.Broker', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('sdv.databroker.v1.Broker', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Broker(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetDatapoints(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdv.databroker.v1.Broker/GetDatapoints',
            sdv_dot_databroker_dot_v1_dot_broker__pb2.GetDatapointsRequest.SerializeToString,
            sdv_dot_databroker_dot_v1_dot_broker__pb2.GetDatapointsReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetDatapoints(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdv.databroker.v1.Broker/SetDatapoints',
            sdv_dot_databroker_dot_v1_dot_broker__pb2.SetDatapointsRequest.SerializeToString,
            sdv_dot_databroker_dot_v1_dot_broker__pb2.SetDatapointsReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Subscribe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/sdv.databroker.v1.Broker/Subscribe',
            sdv_dot_databroker_dot_v1_dot_broker__pb2.SubscribeRequest.SerializeToString,
            sdv_dot_databroker_dot_v1_dot_broker__pb2.SubscribeReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdv.databroker.v1.Broker/GetMetadata',
            sdv_dot_databroker_dot_v1_dot_broker__pb2.GetMetadataRequest.SerializeToString,
            sdv_dot_databroker_dot_v1_dot_broker__pb2.GetMetadataReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



########################################################################
# Copyright (c) 2022-2025 Robert Bosch GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
########################################################################

from __future__ import absolute_import
import contextlib
import dataclasses
import datetime
import enum
import logging
import re
from typing import Any
from typing import Collection
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from pathlib import Path

from google.protobuf import json_format
import grpc
from grpc import RpcError

from kuksa.val.v1 import types_pb2 as types_v1
from kuksa.val.v1 import val_pb2 as val_v1
from kuksa.val.v1 import val_pb2_grpc as val_grpc_v1
from kuksa.val.v2 import types_pb2 as types_v2
from kuksa.val.v2 import val_pb2 as val_v2
from kuksa.val.v2 import val_pb2_grpc as val_grpc_v2

logger = logging.getLogger(__name__)


class DataType(enum.IntEnum):
    UNSPECIFIED = types_v1.DATA_TYPE_UNSPECIFIED
    STRING = types_v1.DATA_TYPE_STRING
    BOOLEAN = types_v1.DATA_TYPE_BOOLEAN
    INT8 = types_v1.DATA_TYPE_INT8
    INT16 = types_v1.DATA_TYPE_INT16
    INT32 = types_v1.DATA_TYPE_INT32
    INT64 = types_v1.DATA_TYPE_INT64
    UINT8 = types_v1.DATA_TYPE_UINT8
    UINT16 = types_v1.DATA_TYPE_UINT16
    UINT32 = types_v1.DATA_TYPE_UINT32
    UINT64 = types_v1.DATA_TYPE_UINT64
    FLOAT = types_v1.DATA_TYPE_FLOAT
    DOUBLE = types_v1.DATA_TYPE_DOUBLE
    TIMESTAMP = types_v1.DATA_TYPE_TIMESTAMP
    STRING_ARRAY = types_v1.DATA_TYPE_STRING_ARRAY
    BOOLEAN_ARRAY = types_v1.DATA_TYPE_BOOLEAN_ARRAY
    INT8_ARRAY = types_v1.DATA_TYPE_INT8_ARRAY
    INT16_ARRAY = types_v1.DATA_TYPE_INT16_ARRAY
    INT32_ARRAY = types_v1.DATA_TYPE_INT32_ARRAY
    INT64_ARRAY = types_v1.DATA_TYPE_INT64_ARRAY
    UINT8_ARRAY = types_v1.DATA_TYPE_UINT8_ARRAY
    UINT16_ARRAY = types_v1.DATA_TYPE_UINT16_ARRAY
    UINT32_ARRAY = types_v1.DATA_TYPE_UINT32_ARRAY
    UINT64_ARRAY = types_v1.DATA_TYPE_UINT64_ARRAY
    FLOAT_ARRAY = types_v1.DATA_TYPE_FLOAT_ARRAY
    DOUBLE_ARRAY = types_v1.DATA_TYPE_DOUBLE_ARRAY
    TIMESTAMP_ARRAY = types_v1.DATA_TYPE_TIMESTAMP_ARRAY


class EntryType(enum.IntEnum):
    UNSPECIFIED = types_v1.ENTRY_TYPE_UNSPECIFIED
    ATTRIBUTE = types_v1.ENTRY_TYPE_ATTRIBUTE
    SENSOR = types_v1.ENTRY_TYPE_SENSOR
    ACTUATOR = types_v1.ENTRY_TYPE_ACTUATOR


class View(enum.IntEnum):
    UNSPECIFIED = types_v1.VIEW_UNSPECIFIED
    CURRENT_VALUE = types_v1.VIEW_CURRENT_VALUE
    TARGET_VALUE = types_v1.VIEW_TARGET_VALUE
    METADATA = types_v1.VIEW_METADATA
    FIELDS = types_v1.VIEW_FIELDS
    ALL = types_v1.VIEW_ALL


class Field(enum.IntEnum):
    UNSPECIFIED = types_v1.FIELD_UNSPECIFIED
    PATH = types_v1.FIELD_PATH
    VALUE = types_v1.FIELD_VALUE
    ACTUATOR_TARGET = types_v1.FIELD_ACTUATOR_TARGET
    METADATA = types_v1.FIELD_METADATA
    METADATA_DATA_TYPE = types_v1.FIELD_METADATA_DATA_TYPE
    METADATA_DESCRIPTION = types_v1.FIELD_METADATA_DESCRIPTION
    METADATA_ENTRY_TYPE = types_v1.FIELD_METADATA_ENTRY_TYPE
    METADATA_COMMENT = types_v1.FIELD_METADATA_COMMENT
    METADATA_DEPRECATION = types_v1.FIELD_METADATA_DEPRECATION
    METADATA_UNIT = types_v1.FIELD_METADATA_UNIT
    METADATA_VALUE_RESTRICTION = types_v1.FIELD_METADATA_VALUE_RESTRICTION
    METADATA_ACTUATOR = types_v1.FIELD_METADATA_ACTUATOR
    METADATA_SENSOR = types_v1.FIELD_METADATA_SENSOR
    METADATA_ATTRIBUTE = types_v1.FIELD_METADATA_ATTRIBUTE


class MetadataField(enum.Enum):
    ALL = Field.METADATA
    DATA_TYPE = Field.METADATA_DATA_TYPE
    DESCRIPTION = Field.METADATA_DESCRIPTION
    ENTRY_TYPE = Field.METADATA_ENTRY_TYPE
    COMMENT = Field.METADATA_COMMENT
    DEPRECATION = Field.METADATA_DEPRECATION
    UNIT = Field.METADATA_UNIT
    VALUE_RESTRICTION = Field.METADATA_VALUE_RESTRICTION
    ACTUATOR = Field.METADATA_ACTUATOR
    SENSOR = Field.METADATA_SENSOR
    ATTRIBUTE = Field.METADATA_ATTRIBUTE


class VSSClientError(Exception):
    def __init__(self, error: Dict[str, Any], errors: List[Dict[str, Any]]):
        super().__init__(error, errors)
        self.error = error
        self.errors = errors

    @classmethod
    def from_grpc_error(cls, error: RpcError):
        grpc_code, grpc_reason = error.code().value
        # TODO: Maybe details could hold an actual Error and/or repeated DataEntryError protobuf messages.
        #       This would allow 'code' to be an actual HTTP/VISS status code not a gRPC one.
        return cls(
            error={
                "code": grpc_code,
                "reason": grpc_reason,
                "message": error.details(),
            },
            errors=[],
        )

    def to_dict(self) -> Dict[str, Any]:
        return {"error": self.error, "errors": self.errors}


@dataclasses.dataclass
class ValueRestriction:
    min: Optional[Any] = None
    max: Optional[Any] = None
    allowed_values: Optional[List[Any]] = None


@dataclasses.dataclass
class Metadata:
    data_type: DataType = DataType.UNSPECIFIED
    entry_type: EntryType = EntryType.UNSPECIFIED
    description: Optional[str] = None
    comment: Optional[str] = None
    deprecation: Optional[str] = None
    unit: Optional[str] = None
    value_restriction: Optional[ValueRestriction] = None
    # No support for entry_specific for now.

    @classmethod
    def from_message(cls, message: types_v1.Metadata):
        metadata = cls(
            data_type=DataType(message.data_type),
            entry_type=EntryType(message.entry_type),
        )
        for field in ("description", "comment", "deprecation", "unit"):
            if message.HasField(field):
                setattr(metadata, field, getattr(message, field))
        if message.HasField("value_restriction"):
            restriction_type = message.value_restriction.WhichOneof("type")
            # Make sure that a type actually is set
            if restriction_type:
                value_restriction = getattr(message.value_restriction, restriction_type)
                metadata.value_restriction = ValueRestriction()
                # All types except string support min/max
                if restriction_type != "string":
                    for field in ("min", "max"):
                        if value_restriction.HasField(field):
                            setattr(
                                metadata.value_restriction,
                                field,
                                getattr(value_restriction, field),
                            )
                if value_restriction.allowed_values:
                    metadata.value_restriction.allowed_values = list(
                        value_restriction.allowed_values
                    )
        return metadata

    # pylint: disable=too-many-branches
    def to_message(
        self, value_type: DataType = DataType.UNSPECIFIED
    ) -> types_v1.Metadata:
        """
        to_message/from_message aligned to use None rather than empty list for
        representing allowed values in value restrictions
        """
        message = types_v1.Metadata(
            data_type=self.data_type.value, entry_type=self.entry_type.value
        )
        for field in ("description", "comment", "deprecation", "unit"):
            field_value = getattr(self, field, None)
            if field_value is not None:
                setattr(message, field, field_value)
        if self.value_restriction is not None:
            if value_type in (
                DataType.INT8,
                DataType.INT16,
                DataType.INT32,
                DataType.INT64,
                DataType.INT8_ARRAY,
                DataType.INT16_ARRAY,
                DataType.INT32_ARRAY,
                DataType.INT64_ARRAY,
            ):
                if self.value_restriction.min is not None:
                    message.value_restriction.signed.min = int(
                        self.value_restriction.min
                    )
                if self.value_restriction.max is not None:
                    message.value_restriction.signed.max = int(
                        self.value_restriction.max
                    )
                if self.value_restriction.allowed_values:
                    message.value_restriction.signed.allowed_values.extend(
                        (int(value) for value in self.value_restriction.allowed_values),
                    )
            elif value_type in (
                DataType.UINT8,
                DataType.UINT16,
                DataType.UINT32,
                DataType.UINT64,
                DataType.UINT8_ARRAY,
                DataType.UINT16_ARRAY,
                DataType.UINT32_ARRAY,
                DataType.UINT64_ARRAY,
            ):
                if self.value_restriction.min is not None:
                    message.value_restriction.unsigned.min = int(
                        self.value_restriction.min
                    )
                if self.value_restriction.max is not None:
                    message.value_restriction.unsigned.max = int(
                        self.value_restriction.max
                    )
                if self.value_restriction.allowed_values:
                    message.value_restriction.unsigned.allowed_values.extend(
                        (int(value) for value in self.value_restriction.allowed_values),
                    )
            elif value_type in (
                DataType.FLOAT,
                DataType.DOUBLE,
                DataType.FLOAT_ARRAY,
                DataType.DOUBLE_ARRAY,
            ):
                if self.value_restriction.min is not None:
                    message.value_restriction.floating_point.min = float(
                        self.value_restriction.min
                    )
                if self.value_restriction.max is not None:
                    message.value_restriction.floating_point.max = float(
                        self.value_restriction.max
                    )
                if self.value_restriction.allowed_values:
                    message.value_restriction.floating_point.allowed_values.extend(
                        (
                            float(value)
                            for value in self.value_restriction.allowed_values
                        ),
                    )
            elif value_type in (
                DataType.STRING,
                DataType.STRING_ARRAY,
            ):
                if self.value_restriction.allowed_values:
                    message.value_restriction.string.allowed_values.extend(
                        (str(value) for value in self.value_restriction.allowed_values),
                    )
            else:
                raise ValueError(
                    f"Cannot set value_restriction from data type {value_type.name}"
                )
        return message

    # pylint: enable=too-many-branches

    @classmethod
    def from_dict(cls, metadata_dict: Dict[str, Any]):
        data_type = metadata_dict.get("data_type", DataType.UNSPECIFIED)
        if isinstance(data_type, str):
            data_type = getattr(DataType, data_type)
        else:
            data_type = DataType(data_type)
        entry_type = metadata_dict.get("entry_type", EntryType.UNSPECIFIED)
        if isinstance(entry_type, str):
            entry_type = getattr(EntryType, entry_type)
        else:
            entry_type = EntryType(entry_type)
        instance = cls(data_type=data_type, entry_type=entry_type)
        for field in ("description", "comment", "deprecation", "unit"):
            field_value = metadata_dict.get(field, None)
            if field_value is not None:
                setattr(instance, field, str(field_value))
        value_restriction = metadata_dict.get("value_restriction")
        if value_restriction is not None:
            instance.value_restriction = ValueRestriction()
            for field in ("min", "max", "allowed_values"):
                field_value = value_restriction.get(field)
                if field_value is not None:
                    setattr(instance.value_restriction, field, field_value)
        return instance

    def to_dict(self) -> Dict[str, Any]:
        out_dict = {
            "data_type": self.data_type.name,
            "entry_type": self.entry_type.name,
        }
        for field in ("description", "comment", "deprecation", "unit"):
            field_value = getattr(self, field, None)
            if field_value is not None:
                out_dict[field] = field_value
        if self.value_restriction is not None:
            out_dict["value_restriction"] = {}
            for field in ("min", "max", "allowed_values"):
                field_value = getattr(self.value_restriction, field, None)
                if field_value is not None:
                    out_dict["value_restriction"][field] = field_value
        return out_dict


@dataclasses.dataclass
class Datapoint:
    value: Optional[Any] = None
    timestamp: Optional[datetime.datetime] = None

    @classmethod
    def from_message(cls, message: types_v1.Datapoint):
        """
        Return internal Datapoint representation or None on error
        """
        if message.WhichOneof("value") is None:
            logger.warning("No value provided in datapoint!")
            return None

        if message.HasField("timestamp"):
            # gRPC timestamp supports date up to including year 9999
            # If timestamp by any reason contains a larger number for seconds than supported
            # you may get an overflow error
            try:
                timestamp = message.timestamp.ToDatetime(
                    tzinfo=datetime.timezone.utc,
                )
            except ValueError:

                logger.error(
                    "Timestamp %d out of accepted range, value ignored!",
                    message.timestamp.seconds,
                )
                return None
        else:
            timestamp = None

        return cls(
            value=getattr(message, message.WhichOneof("value")),
            timestamp=timestamp,
        )

    def cast_array_values(cast, array):
        """
        Parses array input and cast individual values to wanted type.
        Note that input value to this function is not the same as given if you use kuksa-client command line
        as parts (e.g. surrounding quotes) are removed by shell, and then do_setValue also do some magic.
        """
        array = array.strip("[]")

        # Split the input string into separate values
        # First alternative, not quotes including escaped single or double quote, ends at comma, whitespace or EOL
        # Second group is double quoted string, may contain quoted strings and single quotes inside, ends at non-escaped
        # double quote
        # Third is similar but for single quote
        # Using raw strings with surrounding single/double quotes to minimize need for escapes
        pattern = (
            r'(?:\\"|\\' + r"'|[^'" + r'",])+|"(?:\\"|[^"])*"|' + r"'(?:\\'|[^'])*'"
        )
        values = re.findall(pattern, array)
        for item in values:
            # We may in some cases match blanks, that is intended as we want to be able to write arguments like
            # My Way
            # ... without quotes
            if item.strip() == "":
                # skip
                pass
            else:
                yield cast(item)

    def cast_bool(value) -> bool:
        if value in ("False", "false", "F", "f"):
            value = 0
        return bool(value)

    def cast_str(value) -> str:
        """
        Strip based on the following rules.
        - Leading/Trailing blanks are removed
        - If there are single or double quote at both start and end they are removed
        - Finally any quote escapes are removed
        """
        new_val = value.strip()
        if new_val.startswith('"') and new_val.endswith('"'):
            new_val = new_val[1:-1]
        if new_val.startswith("'") and new_val.endswith("'"):
            new_val = new_val[1:-1]
        # Replace escaped quotes with normal quotes
        new_val = new_val.replace('\\"', '"')
        new_val = new_val.replace("\\'", "'")
        return new_val

    def v1_to_message(self, value_type: DataType) -> types_v1.Datapoint:
        message = types_v1.Datapoint()

        def set_array_attr(obj, attr, values):
            array = getattr(obj, attr)
            array.Clear()
            array.values.extend(values)

        field, set_field, cast_field = {
            DataType.INT8: ("int32", setattr, int),
            DataType.INT16: ("int32", setattr, int),
            DataType.INT32: ("int32", setattr, int),
            DataType.UINT8: ("uint32", setattr, int),
            DataType.UINT16: ("uint32", setattr, int),
            DataType.UINT32: ("uint32", setattr, int),
            DataType.UINT64: ("uint64", setattr, int),
            DataType.INT64: ("int64", setattr, int),
            DataType.FLOAT: ("float", setattr, float),
            DataType.DOUBLE: ("double", setattr, float),
            DataType.BOOLEAN: ("bool", setattr, Datapoint.cast_bool),
            DataType.STRING: ("string", setattr, Datapoint.cast_str),
            DataType.INT8_ARRAY: (
                "int32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.INT16_ARRAY: (
                "int32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.INT32_ARRAY: (
                "int32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.UINT8_ARRAY: (
                "uint32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.UINT16_ARRAY: (
                "uint32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.UINT32_ARRAY: (
                "uint32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.UINT64_ARRAY: (
                "uint64_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.INT64_ARRAY: (
                "int64_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.FLOAT_ARRAY: (
                "float_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(float, array),
            ),
            DataType.DOUBLE_ARRAY: (
                "double_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(float, array),
            ),
            DataType.BOOLEAN_ARRAY: (
                "bool_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(Datapoint.cast_bool, array),
            ),
            DataType.STRING_ARRAY: (
                "string_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(Datapoint.cast_str, array),
            ),
        }.get(value_type, (None, None, None))
        if self.value is not None:
            if all((field, set_field, cast_field)):
                set_field(message, field, cast_field(self.value))
            else:
                # Either DataType.TIMESTAMP, DataType.TIMESTAMP_ARRAY or DataType.UNSPECIFIED...
                raise ValueError(
                    f"Cannot determine which field to set with data type {value_type} from value {self.value}",
                )
        if self.timestamp is not None:
            message.timestamp.FromDatetime(self.timestamp)
        return message

    def v2_to_message(self, value_type: DataType) -> types_v2.Datapoint:
        message = types_v2.Datapoint()
        value = types_v2.Value()

        def set_array_attr(obj, attr, values):
            array = getattr(obj, attr)
            array.Clear()
            array.values.extend(values)

        field, set_field, cast_field = {
            DataType.INT8: ("int32", setattr, int),
            DataType.INT16: ("int32", setattr, int),
            DataType.INT32: ("int32", setattr, int),
            DataType.UINT8: ("uint32", setattr, int),
            DataType.UINT16: ("uint32", setattr, int),
            DataType.UINT32: ("uint32", setattr, int),
            DataType.UINT64: ("uint64", setattr, int),
            DataType.INT64: ("int64", setattr, int),
            DataType.FLOAT: ("float", setattr, float),
            DataType.DOUBLE: ("double", setattr, float),
            DataType.BOOLEAN: ("bool", setattr, Datapoint.cast_bool),
            DataType.STRING: ("string", setattr, Datapoint.cast_str),
            DataType.INT8_ARRAY: (
                "int32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.INT16_ARRAY: (
                "int32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.INT32_ARRAY: (
                "int32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.UINT8_ARRAY: (
                "uint32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.UINT16_ARRAY: (
                "uint32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.UINT32_ARRAY: (
                "uint32_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.UINT64_ARRAY: (
                "uint64_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.INT64_ARRAY: (
                "int64_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(int, array),
            ),
            DataType.FLOAT_ARRAY: (
                "float_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(float, array),
            ),
            DataType.DOUBLE_ARRAY: (
                "double_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(float, array),
            ),
            DataType.BOOLEAN_ARRAY: (
                "bool_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(Datapoint.cast_bool, array),
            ),
            DataType.STRING_ARRAY: (
                "string_array",
                set_array_attr,
                lambda array: Datapoint.cast_array_values(Datapoint.cast_str, array),
            ),
        }.get(value_type, (None, None, None))
        if self.value is not None:
            if all((field, set_field, cast_field)):
                set_field(value, field, cast_field(self.value))
                message.value.CopyFrom(value)
            else:
                # Either DataType.TIMESTAMP, DataType.TIMESTAMP_ARRAY or DataType.UNSPECIFIED...
                raise ValueError(
                    f"Cannot determine which field to set with data type {value_type} from value {self.value}",
                )
        if self.timestamp is not None:
            message.timestamp.FromDatetime(self.timestamp)
        return message

    def to_dict(self) -> Dict[str, Any]:
        out_dict = {}
        if self.value is not None:
            out_dict["value"] = self.value
        if self.timestamp is not None:
            out_dict["timestamp"] = self.timestamp.isoformat()
        return out_dict


@dataclasses.dataclass
class DataEntry:
    path: str
    value: Optional[Datapoint] = None
    actuator_target: Optional[Datapoint] = None
    metadata: Optional[Metadata] = None

    # Useful for serialisation, won't appear directly in serialised object
    value_type: DataType = DataType.UNSPECIFIED

    @classmethod
    def from_message(cls, message: types_v1.DataEntry):
        entry_kwargs = {"path": message.path}
        if message.HasField("value"):
            entry_kwargs["value"] = Datapoint.from_message(message.value)
        if message.HasField("actuator_target"):
            entry_kwargs["actuator_target"] = Datapoint.from_message(
                message.actuator_target
            )
        if message.HasField("metadata"):
            entry_kwargs["metadata"] = Metadata.from_message(message.metadata)
        return cls(**entry_kwargs)

    def to_message(self) -> types_v1.DataEntry:
        message = types_v1.DataEntry(path=self.path)
        if self.value is not None:
            message.value.MergeFrom(self.value.v1_to_message(self.value_type))
        if self.actuator_target is not None:
            message.actuator_target.MergeFrom(
                self.actuator_target.v1_to_message(self.value_type)
            )
        if self.metadata is not None:
            message.metadata.MergeFrom(self.metadata.to_message(self.value_type))
        return message

    def to_dict(self) -> Dict[str, Any]:
        out_dict = {"path": self.path}
        if self.value is not None:
            out_dict["value"] = self.value.to_dict()
        if self.actuator_target is not None:
            out_dict["actuator_target"] = self.actuator_target.to_dict()
        if self.metadata is not None:
            out_dict["metadata"] = self.metadata.to_dict()
        return out_dict


@dataclasses.dataclass
class EntryRequest:
    path: str
    view: View
    fields: Iterable[Field]


@dataclasses.dataclass
class EntryUpdate:
    entry: DataEntry
    fields: Iterable[Field]

    @classmethod
    def from_message(cls, message: val_v1.EntryUpdate):
        return cls(
            entry=DataEntry.from_message(message.entry),
            fields=[Field(field) for field in message.fields],
        )

    @classmethod
    def from_tuple(cls, path: str, dp: types_v2.Datapoint):
        # we assume here that at max one field of Value is set -> we use the first entry.
        # If no field is set the value is currently unknown/not avaialable -> set to None.
        data = dp.value.ListFields()
        if data:
            field_descriptor, value = data[0]
            field_name = field_descriptor.name
            value = getattr(dp.value, field_name)
        else:
            value = None
        if dp.timestamp is None or (dp.timestamp.seconds == 0 and dp.timestamp.nanos == 0):
            timestamp = None
        else:
            timestamp = dp.timestamp.ToDatetime(
                tzinfo=datetime.timezone.utc,
            )
        return cls(
            entry=DataEntry(
                path=path, value=Datapoint(value=value, timestamp=timestamp)
            ),
            fields=[Field(value=types_v1.FIELD_VALUE)],
        )

    @classmethod
    def from_actuate_value(cls, path: str, value: types_v2.Value):
        # we assume here that exactly one field of Value is set -> we use the first entry.
        # This should always be the case.
        data = value.ListFields()
        field_descriptor, target_value = data[0]
        field_name = field_descriptor.name
        target_value = getattr(value, field_name)
        return cls(
            entry=DataEntry(
                path=path, actuator_target=Datapoint(value=target_value)
            ),
            fields=[Field(value=types_v1.FIELD_ACTUATOR_TARGET)],
        )

    def to_message(self) -> val_v1.EntryUpdate:
        message = val_v1.EntryUpdate(entry=self.entry.to_message())
        message.fields.extend(field.value for field in self.fields)
        return message

    def to_dict(self) -> Dict[str, Any]:
        return {
            "entry": self.entry.to_dict(),
            "fields": [field.name for field in self.fields],
        }


@dataclasses.dataclass
class SubscribeEntry:
    path: str
    view: View
    fields: Iterable[Field]


@dataclasses.dataclass
class ServerInfo:
    name: str
    version: str

    @classmethod
    def from_message(cls, message: val_v1.GetServerInfoResponse):
        return cls(name=message.name, version=message.version)


class BaseVSSClient:

    def __init__(
        self,
        host: str,
        port: int,
        token: Optional[str] = None,
        root_certificates: Optional[Path] = None,
        ensure_startup_connection: bool = True,
        connected: bool = False,
        tls_server_name: Optional[str] = None,
    ):
        self.authorization_header = self.get_authorization_header(token)
        self.target_host = f"{host}:{port}"
        self.root_certificates = root_certificates
        self.tls_server_name = tls_server_name
        self.ensure_startup_connection = ensure_startup_connection
        self.connected = connected
        self.client_stub_v1 = None
        self.client_stub_v2 = None

    def _load_creds(self) -> Optional[grpc.ChannelCredentials]:
        if self.root_certificates:
            logger.info(f"Using TLS with Root CA from {self.root_certificates}")
            root_certificates = self.root_certificates.read_bytes()
            return grpc.ssl_channel_credentials(root_certificates)
        logger.info(
            "No Root CA present, it will not be possible to use a secure connection!"
        )
        return None

    def _prepare_get_request(
        self, entries: Iterable[EntryRequest]
    ) -> val_v1.GetRequest:
        req = val_v1.GetRequest(entries=[])
        for entry in entries:
            entry_request = val_v1.EntryRequest(
                path=entry.path, view=entry.view.value, fields=[]
            )
            for field in entry.fields:
                entry_request.fields.append(field.value)
            req.entries.append(entry_request)
        logger.debug("%s: %s", type(req).__name__, req)
        return req

    def _process_get_response(self, response: val_v1.GetResponse) -> List[DataEntry]:
        logger.debug("%s: %s", type(response).__name__, response)
        self._raise_if_invalid(response)
        return [DataEntry.from_message(entry) for entry in response.entries]

    def _get_paths_with_required_type(
        self, updates: Collection[EntryUpdate]
    ) -> Dict[str, DataType]:
        paths_with_required_type = {}
        for update in updates:
            metadata = update.entry.metadata
            # We need a data type in order to set sensor/actuator value or metadata's value restriction
            if (
                Field.ACTUATOR_TARGET in update.fields
                or Field.VALUE in update.fields
                or (metadata is not None and metadata.value_restriction is not None)
            ):
                # If the update holds a new data type, we assume it will be applied before
                # setting the sensor/actuator value.
                paths_with_required_type[update.entry.path] = (
                    metadata.data_type if metadata is not None else DataType.UNSPECIFIED
                )
        return paths_with_required_type

    def _prepare_set_request(
        self,
        updates: Collection[EntryUpdate],
        paths_with_required_type: Dict[str, DataType],
    ) -> val_v1.SetRequest:
        req = val_v1.SetRequest(updates=[])
        for update in updates:
            value_type = paths_with_required_type.get(update.entry.path)
            if value_type is not None:
                update.entry.value_type = value_type
            req.updates.append(update.to_message())
        logger.debug("%s: %s", type(req).__name__, req)
        return req

    def _prepare_publish_value_request(
        self,
        update: EntryUpdate,
        paths_with_required_type: Dict[str, DataType],
    ) -> val_v2.PublishValueRequest:
        value_type = paths_with_required_type.get(update.entry.path)
        if value_type is not None:
            update.entry.value_type = value_type

        for field in update.fields:
            if field != Field.VALUE:
                raise VSSClientError(
                    error={
                        "code": grpc.StatusCode.INVALID_ARGUMENT.value[0],
                        "reason": grpc.StatusCode.INVALID_ARGUMENT.value[1],
                        "message": "Cannot use v2 to publish fields other than value",
                    },
                    errors=[],
                )

        req = val_v2.PublishValueRequest(
            signal_id=types_v2.SignalID(path=update.entry.path),
            data_point=update.entry.value.v2_to_message(update.entry.value_type),
        )
        logger.debug("%s: %s", type(req).__name__, req)
        return req

    def _process_set_response(self, response: val_v1.SetResponse) -> None:
        logger.debug("%s: %s", type(response).__name__, response)
        self._raise_if_invalid(response)

    def _prepare_subscribe_request(
        self,
        entries: Iterable[SubscribeEntry],
    ) -> val_v1.SubscribeRequest:
        req = val_v1.SubscribeRequest()
        for entry in entries:
            entry_request = val_v1.SubscribeEntry(
                path=entry.path, view=entry.view.value, fields=[]
            )
            for field in entry.fields:
                entry_request.fields.append(field.value)
            req.entries.append(entry_request)
        logger.debug("%s: %s", type(req).__name__, req)
        return req

    def _prepare_v2_subscribe_request(
        self, paths: Iterable[str]
    ) -> val_v2.SubscribeRequest:
        req = val_v2.SubscribeRequest(signal_paths=paths)
        logger.debug("%s: %s", type(req).__name__, req)
        return req

    def _prepare_v2_provide_actuation_request(
        self,
        paths: Iterable[str],
    ) -> List[val_v2.OpenProviderStreamRequest]:
        signals = []
        for path in paths:
            signals.append(types_v2.SignalID(path=path))
        provide_req = val_v2.ProvideActuationRequest(actuator_identifiers=signals)
        req = val_v2.OpenProviderStreamRequest(provide_actuation_request=provide_req)
        logger.debug("%s: %s", type(req).__name__, req)
        return [req]

    def _prepare_v2_list_metadata_request(
        self, path: str
    ) -> val_v2.ListMetadataRequest:
        req = val_v2.ListMetadataRequest(root=path)
        logger.debug("%s: %s", type(req).__name__, req)
        return req

    def _raise_if_invalid(self, response):
        if response.HasField("error"):
            error = json_format.MessageToDict(
                response.error, preserving_proto_field_name=True
            )
        else:
            error = {}
        if response.errors:
            errors = [
                json_format.MessageToDict(err, preserving_proto_field_name=True)
                for err in response.errors
            ]
        else:
            errors = []

        raise_error = False
        if error and error.get("code") != 200:
            raise_error = True
        else:
            for sub_error in errors:
                if "error" in sub_error:
                    if sub_error["error"].get("code") != 200:
                        logger.debug(
                            "Sub-error %d but no top level error",
                            sub_error["error"].get("code"),
                        )
                        raise_error = True
                else:
                    logger.error("No error field for sub-error")
                    raise_error = True

        if raise_error:
            raise VSSClientError(error, errors)

    def get_authorization_header(self, token: str):
        if token is None:
            return None
        return "Bearer " + token

    def generate_metadata_header(self, metadata: list, header=None) -> list:
        if header is None:
            header = self.authorization_header
        if metadata:
            metadata = dict(metadata)
        else:
            metadata = dict()
        # Overrides old authorization header. This is intentional.
        metadata["authorization"] = header
        return list(metadata.items())


class VSSClient(BaseVSSClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.channel = None
        self.exit_stack = contextlib.ExitStack()
        self.path_to_id_mapping: Dict[str, int] = dict()
        self.id_to_path_mapping: Dict[int, str] = dict()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.disconnect()

    def check_connected(func):
        def wrapper(self, *args, **kwargs):
            if self.connected:
                return func(self, *args, **kwargs)
            else:
                # This shall normally not happen if you use the client as context manager
                # as then a connect will happen automatically when you enter the context
                raise Exception(
                    "Server not connected! Call connect() before using this command!"
                )

        return wrapper

    def connect(self, target_host=None):
        # Reset the id mapping on each new connection to the data broker because the broker
        # could have been restarted and assigned new ids to paths in between.
        # Furthermore, the specified target host could have changed.
        self.path_to_id_mapping.clear()
        self.id_to_path_mapping.clear()

        creds = self._load_creds()
        if target_host is None:
            target_host = self.target_host

        if creds is not None:
            logger.info("Establishing secure channel")
            if self.tls_server_name:
                logger.info(f"Using TLS server name {self.tls_server_name}")
                options = [("grpc.ssl_target_name_override", self.tls_server_name)]
                channel = grpc.secure_channel(target_host, creds, options)
            else:
                logger.debug("Not providing explicit TLS server name")
                channel = grpc.secure_channel(target_host, creds)
        else:
            logger.info("Establishing insecure channel")
            channel = grpc.insecure_channel(target_host)

        self.channel = self.exit_stack.enter_context(channel)
        self.client_stub_v1 = val_grpc_v1.VALStub(self.channel)
        self.client_stub_v2 = val_grpc_v2.VALStub(self.channel)
        self.connected = True
        if self.ensure_startup_connection:
            logger.debug("Connected to server: %s", self.get_server_info())

    def disconnect(self):
        self.exit_stack.close()
        self.client_stub_v1 = None
        self.client_stub_v2 = None
        self.channel = None
        self.connected = False

    @check_connected
    def get_current_values(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> Dict[str, Datapoint]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            current_values = client.get_current_values([
                'Vehicle.Speed',
                'Vehicle.ADAS.ABS.IsActive',
            ])
            speed_value = current_values['Vehicle.Speed'].value
        """
        entries = self.get(
            entries=(
                EntryRequest(path, View.CURRENT_VALUE, (Field.VALUE,)) for path in paths
            ),
            **rpc_kwargs,
        )
        return {entry.path: entry.value for entry in entries}

    @check_connected
    def get_target_values(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> Dict[str, Datapoint]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            target_values = client.get_target_values([
                'Vehicle.ADAS.ABS.IsActive',
            ])
            is_abs_to_become_active = target_values['Vehicle.ADAS.ABS.IsActive'].value
        """
        entries = self.get(
            entries=(
                EntryRequest(
                    path,
                    View.TARGET_VALUE,
                    (Field.ACTUATOR_TARGET,),
                )
                for path in paths
            ),
            **rpc_kwargs,
        )
        return {entry.path: entry.actuator_target for entry in entries}

    @check_connected
    def get_metadata(
        self,
        paths: Iterable[str],
        field: MetadataField = MetadataField.ALL,
        **rpc_kwargs,
    ) -> Dict[str, Metadata]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            metadata = client.get_metadata([
                'Vehicle.Speed',
                'Vehicle.ADAS.ABS.IsActive',
            ], MetadataField.UNIT)
            speed_unit = metadata['Vehicle.Speed'].unit
        """
        entries = self.get(
            entries=(
                EntryRequest(path, View.METADATA, (Field(field.value),))
                for path in paths
            ),
            **rpc_kwargs,
        )
        return {entry.path: entry.metadata for entry in entries}

    @check_connected
    def set_current_values(self, updates: Dict[str, Datapoint], **rpc_kwargs) -> None:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            client.set_current_values({
                'Vehicle.Speed': Datapoint(42),
                'Vehicle.ADAS.ABS.IsActive': Datapoint(False),
            })
        """
        self.set(
            updates=[
                EntryUpdate(DataEntry(path, value=dp), (Field.VALUE,))
                for path, dp in updates.items()
            ],
            try_v2=True,
            **rpc_kwargs,
        )

    @check_connected
    def set_target_values(self, updates: Dict[str, Datapoint], **rpc_kwargs) -> None:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            client.set_target_values(
                {'Vehicle.ADAS.ABS.IsActive': Datapoint(True)})
        """
        self.set(
            updates=[
                EntryUpdate(
                    DataEntry(path, actuator_target=dp),
                    (Field.ACTUATOR_TARGET,),
                )
                for path, dp in updates.items()
            ],
            **rpc_kwargs,
        )

    @check_connected
    def set_metadata(
        self,
        updates: Dict[str, Metadata],
        field: MetadataField = MetadataField.ALL,
        **rpc_kwargs,
    ) -> None:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            client.set_metadata({
                'Vehicle.Cabin.Door.Row1.Left.Shade.Position': Metadata(data_type=DataType.FLOAT),
            })
        """
        self.set(
            updates=[
                EntryUpdate(
                    DataEntry(path, metadata=md),
                    (Field(field.value),),
                )
                for path, md in updates.items()
            ],
            **rpc_kwargs,
        )

    @check_connected
    def subscribe_current_values(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> Iterator[Dict[str, Datapoint]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            for updates in client.subscribe_current_values([
                'Vehicle.Speed', 'Vehicle.ADAS.ABS.IsActive',
            ]):
                for path, dp in updates.items():
                    print(f"Current value for {path} is now: {dp.value}")
        """
        try:
            logger.debug("Try to subscribe current values via v2")
            for updates in self.v2_subscribe(paths, **rpc_kwargs):
                yield {
                    update.entry.path: update.entry.value for update in updates
                }
        except VSSClientError as exc:
            if exc.error["code"] != grpc.StatusCode.UNIMPLEMENTED.value[0]:
                raise

            logger.debug("v2 not available - falling back to v1 subscribe current values")
            for updates in self.subscribe(
                entries=(
                    SubscribeEntry(path, View.CURRENT_VALUE, (Field.VALUE,))
                    for path in paths
                ),
                **rpc_kwargs,
            ):
                yield {update.entry.path: update.entry.value for update in updates}

    @check_connected
    def subscribe_target_values(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> Iterator[Dict[str, Datapoint]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            async for updates in client.subscribe_target_values([
                'Vehicle.ADAS.ABS.IsActive',
            ]):
                for path, dp in updates.items():
                    print(f"Target value for {path} is now: {dp.value}")
        """
        try:
            logger.debug("Try to subscribe actuation requests via v2")
            for updates in self.v2_subscribe_actuation_requests(paths, **rpc_kwargs):
                yield {
                    update.entry.path: update.entry.actuator_target for update in updates
                }
        except VSSClientError as exc:
            if exc.error["code"] != grpc.StatusCode.UNIMPLEMENTED.value[0]:
                raise

            logger.debug("v2 not available - falling back to v1 subscribe target values")
            for updates in self.subscribe(
                entries=(
                    SubscribeEntry(path, View.TARGET_VALUE, (Field.ACTUATOR_TARGET,))
                    for path in paths
                ),
                **rpc_kwargs,
            ):
                yield {
                    update.entry.path: update.entry.actuator_target for update in updates
                }

    @check_connected
    def subscribe_metadata(
        self,
        paths: Iterable[str],
        field: MetadataField = MetadataField.ALL,
        **rpc_kwargs,
    ) -> Iterator[Dict[str, Metadata]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            for updates in client.subscribe_metadata([
                'Vehicle.Speed',
                'Vehicle.ADAS.ABS.IsActive',
            ]):
                for path, md in updates.items():
                    print(f"Metadata for {path} are now: {md.to_dict()}")
        """
        for updates in self.subscribe(
            entries=(
                SubscribeEntry(path, View.METADATA, (Field(field.value),))
                for path in paths
            ),
            **rpc_kwargs,
        ):
            yield {update.entry.path: update.entry.metadata for update in updates}

    @check_connected
    def get(self, entries: Iterable[EntryRequest], **rpc_kwargs) -> List[DataEntry]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        req = self._prepare_get_request(entries)
        try:
            resp = self.client_stub_v1.Get(req, **rpc_kwargs)
        except RpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc
        return self._process_get_response(resp)

    @check_connected
    def set(
        self, updates: Collection[EntryUpdate], try_v2: bool = False, **rpc_kwargs
    ) -> None:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        paths_with_required_type = self._get_paths_with_required_type(updates)
        paths_without_type = [
            path
            for path, data_type in paths_with_required_type.items()
            if data_type is DataType.UNSPECIFIED
        ]
        paths_with_required_type.update(
            self.get_value_types(paths_without_type, **rpc_kwargs)
        )
        if try_v2:
            logger.debug("Trying v2")
            if len(updates) == 0:
                raise VSSClientError(
                    error={
                        "code": grpc.StatusCode.INVALID_ARGUMENT.value[0],
                        "reason": grpc.StatusCode.INVALID_ARGUMENT.value[1],
                        "message": "No datapoints requested",
                    },
                    errors=[],
                )
            for update in updates:
                req = self._prepare_publish_value_request(
                    update, paths_with_required_type
                )
                try:
                    resp = self.client_stub_v2.PublishValue(req, **rpc_kwargs)
                except RpcError as exc:
                    if exc.code() == grpc.StatusCode.UNIMPLEMENTED:
                        logger.debug("v2 not available fall back to v1 instead")
                        self.set(updates)
                    else:
                        raise VSSClientError.from_grpc_error(exc) from exc
        else:
            logger.debug("Trying v1")
            req = self._prepare_set_request(updates, paths_with_required_type)
            try:
                resp = self.client_stub_v1.Set(req, **rpc_kwargs)
            except RpcError as exc:
                raise VSSClientError.from_grpc_error(exc) from exc
            self._process_set_response(resp)

    def get_path(self, signal_id: types_v2.SignalID) -> str:
        if signal_id.HasField("path"):
            return signal_id.path
        elif signal_id.HasField("id") and signal_id.id in self.id_to_path_mapping:
            return self.id_to_path_mapping[signal_id.id]
        return "<unknown signal>"

    def ensure_id_mapping(self, paths: Iterable[str], **rpc_kwargs):
        for path in paths:
            if path not in self.path_to_id_mapping:
                # Prevent duplicate requests for the same path
                self.path_to_id_mapping[path] = None
                req = self._prepare_v2_list_metadata_request(path)
                try:
                    resp = self.client_stub_v2.ListMetadata(req, **rpc_kwargs)
                    logger.debug("%s: %s", type(resp).__name__, resp)
                    if len(resp.metadata) == 1:
                        self.path_to_id_mapping[path] = resp.metadata[0].id
                        self.id_to_path_mapping[resp.metadata[0].id] = path
                    else:
                        del self.path_to_id_mapping[path]
                        raise VSSClientError(
                            error={
                                "code": grpc.StatusCode.NOT_FOUND.value[0],
                                "reason": grpc.StatusCode.NOT_FOUND.value[1],
                                "message": f"Path {path} not found on server",
                            },
                            errors=[],
                        )
                except RpcError as exc:
                    if exc.code() == grpc.StatusCode.UNIMPLEMENTED:
                        logger.debug("v2 not available - skip querying ids")
                        return
                    raise VSSClientError.from_grpc_error(exc) from exc

    @check_connected
    def subscribe(
        self, entries: Iterable[SubscribeEntry], try_v2: bool = False, **rpc_kwargs
    ) -> Iterator[List[EntryUpdate]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """

        if try_v2:
            raise VSSClientError(
                error={
                    "code": grpc.StatusCode.INVALID_ARGUMENT.value[0],
                    "reason": grpc.StatusCode.INVALID_ARGUMENT.value[1],
                    "message": ("Method subscribe supports v1, only. "
                                "Use v2_subscribe or v2_subscribe_actuation_requests instead."),
                },
                errors=[],
            )

        logger.debug("Try subscribing via v1")
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        req = self._prepare_subscribe_request(entries)
        resp_stream = self.client_stub_v1.Subscribe(req, **rpc_kwargs)
        try:
            for resp in resp_stream:
                logger.debug("%s: %s", type(resp).__name__, resp)
                yield [EntryUpdate.from_message(update) for update in resp.updates]
        except RpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc

    @check_connected
    def v2_subscribe(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> Iterator[List[EntryUpdate]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """

        logger.debug("Subscribe current values via v2")
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        req = self._prepare_v2_subscribe_request(paths)
        resp_stream = self.client_stub_v2.Subscribe(req, **rpc_kwargs)
        try:
            for resp in resp_stream:
                logger.debug("%s: %s", type(resp).__name__, resp)
                yield [
                    EntryUpdate.from_tuple(path, dp)
                    for path, dp in resp.entries.items()
                ]
        except RpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc

    @check_connected
    def v2_subscribe_actuation_requests(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> Iterator[List[EntryUpdate]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """

        logger.debug("Subscribe actuation requests via v2")
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        self.ensure_id_mapping(paths, **rpc_kwargs)
        req = self._prepare_v2_provide_actuation_request(paths)
        resp_stream = self.client_stub_v2.OpenProviderStream(iter(req), **rpc_kwargs)
        try:
            for resp in resp_stream:
                logger.debug("batch %s: %s", type(resp).__name__, resp)
                if resp.HasField("batch_actuate_stream_request"):
                    yield [
                        EntryUpdate.from_actuate_value(self.get_path(actuate_req.signal_id), actuate_req.value)
                        for actuate_req in resp.batch_actuate_stream_request.actuate_requests
                    ]
        except RpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc

    @check_connected
    def authorize(self, token: str, **rpc_kwargs) -> str:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
            token
                string containing the actual token
        """
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            metadata=rpc_kwargs.get("metadata"),
            header=self.get_authorization_header(token),
        )
        req = val_v1.GetServerInfoRequest()
        try:
            resp = self.client_stub_v1.GetServerInfo(req, **rpc_kwargs)
        except RpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc
        logger.debug("%s: %s", type(resp).__name__, resp)
        self.authorization_header = self.get_authorization_header(token)
        return "Authenticated"

    @check_connected
    def get_server_info(self, **rpc_kwargs) -> Optional[ServerInfo]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            metadata=rpc_kwargs.get("metadata")
        )
        req = val_v1.GetServerInfoRequest()
        logger.debug("%s: %s", type(req).__name__, req)
        try:
            resp = self.client_stub_v1.GetServerInfo(req, **rpc_kwargs)
            logger.debug("%s: %s", type(resp).__name__, resp)
            return ServerInfo.from_message(resp)
        except RpcError as exc:
            if exc.code() == grpc.StatusCode.UNAUTHENTICATED:
                logger.info("Unauthenticated channel started")
            else:
                raise VSSClientError.from_grpc_error(exc) from exc
        return None

    @check_connected
    def get_value_types(
        self, paths: Collection[str], **rpc_kwargs
    ) -> Dict[str, DataType]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        req = self._prepare_get_request(entries)
        """
        if paths:
            entry_requests = (
                EntryRequest(
                    path=path,
                    view=View.METADATA,
                    fields=(Field.METADATA_DATA_TYPE,),
                )
                for path in paths
            )
            entries = self.get(entries=entry_requests, **rpc_kwargs)
            return {entry.path: DataType(entry.metadata.data_type) for entry in entries}
        return {}



########################################################################
# Copyright (c) 2022-2025 Robert Bosch GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
########################################################################

import pathlib

import grpc.aio
import pytest
import pytest_asyncio

from kuksa.val.v1 import val_pb2_grpc as val_v1
from kuksa.val.v2 import val_pb2_grpc as val_v2

import tests


@pytest.fixture(name='resources_path', scope='function')
def resources_path_fixture():
    return pathlib.Path(tests.__path__[0]) / 'resources'


@pytest.fixture(name="val_servicer_v1", scope="function")
def val_servicer_v1_fixture(mocker):
    servicer_v1 = val_v1.VALServicer()
    mocker.patch.object(servicer_v1, "Get", spec=True)
    mocker.patch.object(servicer_v1, "Set", spec=True)
    mocker.patch.object(servicer_v1, "Subscribe", spec=True)
    mocker.patch.object(servicer_v1, "GetServerInfo", spec=True)

    return servicer_v1


@pytest.fixture(name="val_servicer_v2", scope="function")
def val_servicer_v2_fixture(mocker):
    servicer_v2 = val_v2.VALServicer()
    mocker.patch.object(servicer_v2, "ListMetadata", spec=True)
    mocker.patch.object(servicer_v2, "OpenProviderStream", spec=True)
    mocker.patch.object(servicer_v2, "PublishValue", spec=True)
    mocker.patch.object(servicer_v2, "Subscribe", spec=True)

    return servicer_v2


@pytest_asyncio.fixture(name="mocked_databroker", scope="function")
async def val_server_fixture(unused_tcp_port, val_servicer_v1, val_servicer_v2):
    server = grpc.aio.server()
    val_v1.add_VALServicer_to_server(val_servicer_v1, server)
    val_v2.add_VALServicer_to_server(val_servicer_v2, server)
    server.add_insecure_port(f"127.0.0.1:{unused_tcp_port}")
    await server.start()
    try:
        yield server
    finally:
        await server.stop(grace=2.0)


@pytest_asyncio.fixture(name="secure_mocked_databroker", scope="function")
async def secure_val_server_fixture(
    unused_tcp_port, resources_path, val_servicer_v1, val_servicer_v2
):
    server = grpc.aio.server()
    val_v1.add_VALServicer_to_server(val_servicer_v1, server)
    val_v2.add_VALServicer_to_server(val_servicer_v2, server)
    server.add_secure_port(
        f"localhost:{unused_tcp_port}",
        grpc.ssl_server_credentials(
            private_key_certificate_chain_pairs=[
                (
                    (resources_path / "test-server.key").read_bytes(),
                    (resources_path / "test-server.pem").read_bytes(),
                )
            ],
            root_certificates=(resources_path / "test-ca.pem").read_bytes(),
            require_client_auth=False,
        ),
    )
    await server.start()
    try:
        yield server
    finally:
        await server.stop(grace=2.0)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: sdv/databroker/v1/broker.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'sdv/databroker/v1/broker.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sdv.databroker.v1 import types_pb2 as sdv_dot_databroker_dot_v1_dot_types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1esdv/databroker/v1/broker.proto\x12\x11sdv.databroker.v1\x1a\x1dsdv/databroker/v1/types.proto\"*\n\x14GetDatapointsRequest\x12\x12\n\ndatapoints\x18\x01 \x03(\t\"\xb0\x01\n\x12GetDatapointsReply\x12I\n\ndatapoints\x18\x01 \x03(\x0b\x32\x35.sdv.databroker.v1.GetDatapointsReply.DatapointsEntry\x1aO\n\x0f\x44\x61tapointsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.sdv.databroker.v1.Datapoint:\x02\x38\x01\"\xb4\x01\n\x14SetDatapointsRequest\x12K\n\ndatapoints\x18\x01 \x03(\x0b\x32\x37.sdv.databroker.v1.SetDatapointsRequest.DatapointsEntry\x1aO\n\x0f\x44\x61tapointsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.sdv.databroker.v1.Datapoint:\x02\x38\x01\"\xa9\x01\n\x12SetDatapointsReply\x12\x41\n\x06\x65rrors\x18\x01 \x03(\x0b\x32\x31.sdv.databroker.v1.SetDatapointsReply.ErrorsEntry\x1aP\n\x0b\x45rrorsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x30\n\x05value\x18\x02 \x01(\x0e\x32!.sdv.databroker.v1.DatapointError:\x02\x38\x01\"!\n\x10SubscribeRequest\x12\r\n\x05query\x18\x02 \x01(\t\"\x9c\x01\n\x0eSubscribeReply\x12=\n\x06\x66ields\x18\x01 \x03(\x0b\x32-.sdv.databroker.v1.SubscribeReply.FieldsEntry\x1aK\n\x0b\x46ieldsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.sdv.databroker.v1.Datapoint:\x02\x38\x01\"#\n\x12GetMetadataRequest\x12\r\n\x05names\x18\x01 \x03(\t\"=\n\x10GetMetadataReply\x12)\n\x04list\x18\x01 \x03(\x0b\x32\x1b.sdv.databroker.v1.Metadata2\xfc\x02\n\x06\x42roker\x12_\n\rGetDatapoints\x12\'.sdv.databroker.v1.GetDatapointsRequest\x1a%.sdv.databroker.v1.GetDatapointsReply\x12_\n\rSetDatapoints\x12\'.sdv.databroker.v1.SetDatapointsRequest\x1a%.sdv.databroker.v1.SetDatapointsReply\x12U\n\tSubscribe\x12#.sdv.databroker.v1.SubscribeRequest\x1a!.sdv.databroker.v1.SubscribeReply0\x01\x12Y\n\x0bGetMetadata\x12%.sdv.databroker.v1.GetMetadataRequest\x1a#.sdv.databroker.v1.GetMetadataReplyb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sdv.databroker.v1.broker_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GETDATAPOINTSREPLY_DATAPOINTSENTRY']._loaded_options = None
  _globals['_GETDATAPOINTSREPLY_DATAPOINTSENTRY']._serialized_options = b'8\001'
  _globals['_SETDATAPOINTSREQUEST_DATAPOINTSENTRY']._loaded_options = None
  _globals['_SETDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_options = b'8\001'
  _globals['_SETDATAPOINTSREPLY_ERRORSENTRY']._loaded_options = None
  _globals['_SETDATAPOINTSREPLY_ERRORSENTRY']._serialized_options = b'8\001'
  _globals['_SUBSCRIBEREPLY_FIELDSENTRY']._loaded_options = None
  _globals['_SUBSCRIBEREPLY_FIELDSENTRY']._serialized_options = b'8\001'
  _globals['_GETDATAPOINTSREQUEST']._serialized_start=84
  _globals['_GETDATAPOINTSREQUEST']._serialized_end=126
  _globals['_GETDATAPOINTSREPLY']._serialized_start=129
  _globals['_GETDATAPOINTSREPLY']._serialized_end=305
  _globals['_GETDATAPOINTSREPLY_DATAPOINTSENTRY']._serialized_start=226
  _globals['_GETDATAPOINTSREPLY_DATAPOINTSENTRY']._serialized_end=305
  _globals['_SETDATAPOINTSREQUEST']._serialized_start=308
  _globals['_SETDATAPOINTSREQUEST']._serialized_end=488
  _globals['_SETDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_start=226
  _globals['_SETDATAPOINTSREQUEST_DATAPOINTSENTRY']._serialized_end=305
  _globals['_SETDATAPOINTSREPLY']._serialized_start=491
  _globals['_SETDATAPOINTSREPLY']._serialized_end=660
  _globals['_SETDATAPOINTSREPLY_ERRORSENTRY']._serialized_start=580
  _globals['_SETDATAPOINTSREPLY_ERRORSENTRY']._serialized_end=660
  _globals['_SUBSCRIBEREQUEST']._serialized_start=662
  _globals['_SUBSCRIBEREQUEST']._serialized_end=695
  _globals['_SUBSCRIBEREPLY']._serialized_start=698
  _globals['_SUBSCRIBEREPLY']._serialized_end=854
  _globals['_SUBSCRIBEREPLY_FIELDSENTRY']._serialized_start=779
  _globals['_SUBSCRIBEREPLY_FIELDSENTRY']._serialized_end=854
  _globals['_GETMETADATAREQUEST']._serialized_start=856
  _globals['_GETMETADATAREQUEST']._serialized_end=891
  _globals['_GETMETADATAREPLY']._serialized_start=893
  _globals['_GETMETADATAREPLY']._serialized_end=954
  _globals['_BROKER']._serialized_start=957
  _globals['_BROKER']._serialized_end=1337
# @@protoc_insertion_point(module_scope)



########################################################################
# Copyright (c) 2022-2025 Robert Bosch GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
########################################################################

import asyncio
import contextlib
import logging
from typing import AsyncIterator
from typing import Callable
from typing import Collection
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
import uuid

import grpc
from grpc.aio import AioRpcError

from kuksa.val.v1 import val_pb2 as val_v1
from kuksa.val.v1 import val_pb2_grpc as val_grpc_v1
from kuksa.val.v2 import types_pb2 as types_v2
from kuksa.val.v2 import val_pb2_grpc as val_grpc_v2

from . import BaseVSSClient
from . import Datapoint
from . import DataEntry
from . import DataType
from . import EntryRequest
from . import EntryUpdate
from . import Field
from . import Metadata
from . import MetadataField
from . import ServerInfo
from . import SubscribeEntry
from . import View
from . import VSSClientError

logger = logging.getLogger(__name__)


class VSSClient(BaseVSSClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.channel = None
        self.exit_stack = contextlib.AsyncExitStack()
        self.path_to_id_mapping: Dict[str, int] = dict()
        self.id_to_path_mapping: Dict[int, str] = dict()

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.disconnect()

    async def connect(self, target_host=None):
        self.path_to_id_mapping.clear()
        self.id_to_path_mapping.clear()

        creds = self._load_creds()
        if target_host is None:
            target_host = self.target_host

        if creds is not None:
            logger.info("Establishing secure channel")
            if self.tls_server_name:
                logger.info(f"Using TLS server name {self.tls_server_name}")
                options = [("grpc.ssl_target_name_override", self.tls_server_name)]
                channel = grpc.aio.secure_channel(target_host, creds, options)
            else:
                logger.debug("Not providing explicit TLS server name")
                channel = grpc.aio.secure_channel(target_host, creds)
        else:
            logger.info("Establishing insecure channel")
            channel = grpc.aio.insecure_channel(target_host)

        self.channel = await self.exit_stack.enter_async_context(channel)
        self.client_stub_v1 = val_grpc_v1.VALStub(self.channel)
        self.client_stub_v2 = val_grpc_v2.VALStub(self.channel)
        self.connected = True
        if self.ensure_startup_connection:
            logger.debug("Connected to server: %s", await self.get_server_info())

    async def disconnect(self):
        await self.exit_stack.aclose()
        self.client_stub_v1 = None
        self.client_stub_v2 = None
        self.channel = None
        self.connected = False

    def check_connected_async(func):
        """
        Decorator to verify that there is a connection before calling underlying method
        For generator methods use check_connected_async_iter
        """

        async def wrapper(self, *args, **kwargs):
            if self.connected:
                return await func(self, *args, **kwargs)
            else:
                # This shall normally not happen if you use the client as context manager
                # as then a connect will happen automatically when you enter the context
                raise Exception(
                    "Server not connected! Call connect() before using this command!"
                )

        return wrapper

    def check_connected_async_iter(func):
        """
        Decorator for generator methods to verify that there is a connection before calling underlying method
        """

        async def wrapper(self, *args, **kwargs):
            if self.connected:
                async for v in func(self, *args, **kwargs):
                    yield v
            else:
                # This shall normally not happen if you use the client as context manager
                # as then a connect will happen automatically when you enter the context
                raise Exception(
                    "Server not connected! Call connect() before using this command!"
                )

        return wrapper

    @check_connected_async
    async def get_current_values(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> Dict[str, Datapoint]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            current_values = await client.get_current_values([
                'Vehicle.Speed',
                'Vehicle.ADAS.ABS.IsActive',
            ])
            speed_value = current_values['Vehicle.Speed'].value
        """
        entries = await self.get(
            entries=(
                EntryRequest(path, View.CURRENT_VALUE, (Field.VALUE,)) for path in paths
            ),
            **rpc_kwargs,
        )
        return {entry.path: entry.value for entry in entries}

    @check_connected_async
    async def get_target_values(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> Dict[str, Datapoint]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            target_values = await client.get_target_values([
                'Vehicle.ADAS.ABS.IsActive',
            ])
            is_abs_to_become_active = target_values['Vehicle.ADAS.ABS.IsActive'].value
        """
        entries = await self.get(
            entries=(
                EntryRequest(
                    path,
                    View.TARGET_VALUE,
                    (Field.ACTUATOR_TARGET,),
                    **rpc_kwargs,
                )
                for path in paths
            )
        )
        return {entry.path: entry.actuator_target for entry in entries}

    @check_connected_async
    async def get_metadata(
        self,
        paths: Iterable[str],
        field: MetadataField = MetadataField.ALL,
        **rpc_kwargs,
    ) -> Dict[str, Metadata]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            metadata = await client.get_metadata([
                'Vehicle.Speed',
                'Vehicle.ADAS.ABS.IsActive',
            ], MetadataField.UNIT)
            speed_unit = metadata['Vehicle.Speed'].unit
        """
        entries = await self.get(
            entries=(
                EntryRequest(path, View.METADATA, (Field(field.value),))
                for path in paths
            ),
            **rpc_kwargs,
        )
        return {entry.path: entry.metadata for entry in entries}

    @check_connected_async
    async def set_current_values(
        self, updates: Dict[str, Datapoint], **rpc_kwargs
    ) -> None:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            await client.set_current_values({
                'Vehicle.Speed': Datapoint(42),
                'Vehicle.ADAS.ABS.IsActive': Datapoint(False),
            })
        """
        logger.info("Setting current value")
        await self.set(
            updates=[
                EntryUpdate(DataEntry(path, value=dp), (Field.VALUE,))
                for path, dp in updates.items()
            ],
            try_v2=True,
            **rpc_kwargs,
        )

    @check_connected_async
    async def set_target_values(
        self, updates: Dict[str, Datapoint], **rpc_kwargs
    ) -> None:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            await client.set_target_values({'Vehicle.ADAS.ABS.IsActive': Datapoint(True)})
        """
        await self.set(
            updates=[
                EntryUpdate(
                    DataEntry(path, actuator_target=dp),
                    (Field.ACTUATOR_TARGET,),
                )
                for path, dp in updates.items()
            ],
            **rpc_kwargs,
        )

    @check_connected_async
    async def set_metadata(
        self,
        updates: Dict[str, Metadata],
        field: MetadataField = MetadataField.ALL,
        **rpc_kwargs,
    ) -> None:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            await client.set_metadata({
                'Vehicle.Cabin.Door.Row1.Left.Shade.Position': Metadata(data_type=DataType.FLOAT),
            })
        """
        await self.set(
            updates=[
                EntryUpdate(
                    DataEntry(path, metadata=md),
                    (Field(field.value),),
                )
                for path, md in updates.items()
            ],
            **rpc_kwargs,
        )

    @check_connected_async_iter
    async def subscribe_current_values(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> AsyncIterator[Dict[str, Datapoint]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            async for updates in client.subscribe_current_values([
                'Vehicle.Speed', 'Vehicle.ADAS.ABS.IsActive',
            ]):
                for path, dp in updates.items():
                    print(f"Current value for {path} is now: {dp.value}")
        """
        try:
            logger.debug("Try to subscribe current values via v2")
            async for updates in self.v2_subscribe(paths=paths, **rpc_kwargs):
                yield {
                    update.entry.path: update.entry.value for update in updates
                }
        except VSSClientError as exc:
            if exc.error["code"] != grpc.StatusCode.UNIMPLEMENTED.value[0]:
                raise

            logger.debug("v2 not available - falling back to v1 subscribe current values")
            async for updates in self.subscribe(
                entries=(
                    SubscribeEntry(path, View.CURRENT_VALUE, (Field.VALUE,))
                    for path in paths
                ),
                **rpc_kwargs,
            ):
                yield {update.entry.path: update.entry.value for update in updates}

    @check_connected_async_iter
    async def subscribe_target_values(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> AsyncIterator[Dict[str, Datapoint]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            async for updates in client.subscribe_target_values([
                'Vehicle.ADAS.ABS.IsActive',
            ]):
                for path, dp in updates.items():
                    print(f"Target value for {path} is now: {dp.value}")
        """
        try:
            logger.debug("Try to subscribe actuation requests via v2")
            async for updates in self.v2_subscribe_actuation_requests(paths=paths, **rpc_kwargs):
                yield {
                    update.entry.path: update.entry.actuator_target for update in updates
                }
        except VSSClientError as exc:
            if exc.error["code"] != grpc.StatusCode.UNIMPLEMENTED.value[0]:
                raise

            logger.debug("v2 not available - falling back to v1 subscribe target values")
            async for updates in self.subscribe(
                entries=(
                    SubscribeEntry(path, View.TARGET_VALUE, (Field.ACTUATOR_TARGET,))
                    for path in paths
                ),
                **rpc_kwargs,
            ):
                yield {
                    update.entry.path: update.entry.actuator_target for update in updates
                }

    @check_connected_async_iter
    async def subscribe_metadata(
        self,
        paths: Iterable[str],
        field: MetadataField = MetadataField.ALL,
        **rpc_kwargs,
    ) -> AsyncIterator[Dict[str, Metadata]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        Example:
            async for updates in client.subscribe_metadata([
                'Vehicle.Speed',
                'Vehicle.ADAS.ABS.IsActive',
            ]):
                for path, md in updates.items():
                    print(f"Metadata for {path} are now: {md.to_dict()}")
        """
        async for updates in self.subscribe(
            entries=(
                SubscribeEntry(path, View.METADATA, (Field(field.value),))
                for path in paths
            ),
            **rpc_kwargs,
        ):
            yield {update.entry.path: update.entry.metadata for update in updates}

    @check_connected_async
    async def get(
        self, entries: Iterable[EntryRequest], **rpc_kwargs
    ) -> List[DataEntry]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        req = self._prepare_get_request(entries)
        try:
            resp = await self.client_stub_v1.Get(req, **rpc_kwargs)
        except AioRpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc
        return self._process_get_response(resp)

    @check_connected_async
    async def set(
        self, updates: Collection[EntryUpdate], try_v2: bool = False, **rpc_kwargs
    ) -> None:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        paths_with_required_type = self._get_paths_with_required_type(updates)
        paths_without_type = [
            path
            for path, data_type in paths_with_required_type.items()
            if data_type is DataType.UNSPECIFIED
        ]
        paths_with_required_type.update(
            await self.get_value_types(paths_without_type, **rpc_kwargs)
        )
        if try_v2:
            logger.debug("Trying v2")
            if len(updates) == 0:
                raise VSSClientError(
                    error={
                        "code": grpc.StatusCode.INVALID_ARGUMENT.value[0],
                        "reason": grpc.StatusCode.INVALID_ARGUMENT.value[1],
                        "message": "No datapoints requested",
                    },
                    errors=[],
                )
            for update in updates:
                req = self._prepare_publish_value_request(
                    update, paths_with_required_type
                )
                try:
                    resp = await self.client_stub_v2.PublishValue(req, **rpc_kwargs)
                except AioRpcError as exc:
                    if exc.code() == grpc.StatusCode.UNIMPLEMENTED:
                        logger.debug("v2 not available fall back to v1 instead")
                        await self.set(updates)
                    else:
                        raise VSSClientError.from_grpc_error(exc) from exc
        else:
            logger.debug("Trying v1")
            req = self._prepare_set_request(updates, paths_with_required_type)
            try:
                resp = await self.client_stub_v1.Set(req, **rpc_kwargs)
            except AioRpcError as exc:
                raise VSSClientError.from_grpc_error(exc) from exc
            self._process_set_response(resp)

    def get_path(self, signal_id: types_v2.SignalID) -> str:
        if signal_id.HasField("path"):
            return signal_id.path
        elif signal_id.HasField("id") and signal_id.id in self.id_to_path_mapping:
            return self.id_to_path_mapping[signal_id.id]
        return "<unknown signal>"

    async def ensure_id_mapping(self, paths: Iterable[str], **rpc_kwargs):
        for path in paths:
            if path not in self.path_to_id_mapping:
                # Prevent duplicate requests for the same path
                self.path_to_id_mapping[path] = None
                req = self._prepare_v2_list_metadata_request(path)
                try:
                    resp = await self.client_stub_v2.ListMetadata(req, **rpc_kwargs)
                    logger.debug("%s: %s", type(resp).__name__, resp)
                    if len(resp.metadata) == 1:
                        self.path_to_id_mapping[path] = resp.metadata[0].id
                        self.id_to_path_mapping[resp.metadata[0].id] = path
                    else:
                        del self.path_to_id_mapping[path]
                        raise VSSClientError(
                            error={
                                "code": grpc.StatusCode.NOT_FOUND.value[0],
                                "reason": grpc.StatusCode.NOT_FOUND.value[1],
                                "message": f"Path {path} not found on server",
                            },
                            errors=[],
                        )
                except AioRpcError as exc:
                    if exc.code() == grpc.StatusCode.UNIMPLEMENTED:
                        logger.debug("v2 not available - skip querying ids")
                        return
                    raise VSSClientError.from_grpc_error(exc) from exc

    @check_connected_async_iter
    async def subscribe(
        self,
        entries: Iterable[SubscribeEntry],
        try_v2: bool = False,
        **rpc_kwargs,
    ) -> AsyncIterator[List[EntryUpdate]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """

        if try_v2:
            raise VSSClientError(
                error={
                    "code": grpc.StatusCode.INVALID_ARGUMENT.value[0],
                    "reason": grpc.StatusCode.INVALID_ARGUMENT.value[1],
                    "message": ("Method subscribe supports v1, only. "
                                "Use v2_subscribe or v2_subscribe_actuation_requests instead."),
                },
                errors=[],
            )

        logger.debug("Try subscribing via v1")
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        req = self._prepare_subscribe_request(entries)
        resp_stream = self.client_stub_v1.Subscribe(req, **rpc_kwargs)
        try:
            async for resp in resp_stream:
                logger.debug("%s: %s", type(resp).__name__, resp)
                yield [EntryUpdate.from_message(update) for update in resp.updates]
        except AioRpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc

    @check_connected_async_iter
    async def v2_subscribe(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> AsyncIterator[List[EntryUpdate]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """

        logger.debug("Subscribe current values via v2")
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        req = self._prepare_v2_subscribe_request(paths)
        resp_stream = self.client_stub_v2.Subscribe(req, **rpc_kwargs)
        try:
            async for resp in resp_stream:
                logger.debug("%s: %s", type(resp).__name__, resp)
                yield [
                    EntryUpdate.from_tuple(path, dp)
                    for path, dp in resp.entries.items()
                ]
        except AioRpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc

    @check_connected_async_iter
    async def v2_subscribe_actuation_requests(
        self, paths: Iterable[str], **rpc_kwargs
    ) -> AsyncIterator[List[EntryUpdate]]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """

        logger.debug("Subscribe actuation requests via v2")
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            rpc_kwargs.get("metadata")
        )
        await self.ensure_id_mapping(paths, **rpc_kwargs)
        req = self._prepare_v2_provide_actuation_request(paths)
        resp_stream = self.client_stub_v2.OpenProviderStream(iter(req), **rpc_kwargs)
        try:
            async for resp in resp_stream:
                logger.debug("batch %s: %s", type(resp).__name__, resp)
                if resp.HasField("batch_actuate_stream_request"):
                    yield [
                        EntryUpdate.from_actuate_value(self.get_path(actuate_req.signal_id), actuate_req.value)
                        for actuate_req in resp.batch_actuate_stream_request.actuate_requests
                    ]
        except AioRpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc

    @check_connected_async
    async def authorize(self, token: str, **rpc_kwargs) -> str:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
            token
                string containing the actual token
        """
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            metadata=rpc_kwargs.get("metadata"),
            header=self.get_authorization_header(token),
        )
        req = val_v1.GetServerInfoRequest()
        try:
            resp = await self.client_stub_v1.GetServerInfo(req, **rpc_kwargs)
        except AioRpcError as exc:
            raise VSSClientError.from_grpc_error(exc) from exc
        logger.debug("%s: %s", type(resp).__name__, resp)
        self.authorization_header = self.get_authorization_header(token)
        return "Authenticated"

    @check_connected_async
    async def get_server_info(self, **rpc_kwargs) -> Optional[ServerInfo]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """
        rpc_kwargs["metadata"] = self.generate_metadata_header(
            metadata=rpc_kwargs.get("metadata")
        )
        req = val_v1.GetServerInfoRequest()
        logger.debug("%s: %s", type(req).__name__, req)
        try:
            resp = await self.client_stub_v1.GetServerInfo(req, **rpc_kwargs)
            logger.debug("%s: %s", type(resp).__name__, resp)
            return ServerInfo.from_message(resp)
        except AioRpcError as exc:
            if exc.code() == grpc.StatusCode.UNAUTHENTICATED:
                logger.info("Unauthenticated channel started")
            else:
                raise VSSClientError.from_grpc_error(exc) from exc
        return None

    @check_connected_async
    async def get_value_types(
        self, paths: Collection[str], **rpc_kwargs
    ) -> Dict[str, DataType]:
        """
        Parameters:
            rpc_kwargs
                grpc.*MultiCallable kwargs e.g. timeout, metadata, credentials.
        """
        if paths:
            entry_requests = (
                EntryRequest(
                    path=path,
                    view=View.METADATA,
                    fields=(Field.METADATA_DATA_TYPE,),
                )
                for path in paths
            )
            entries = await self.get(entries=entry_requests, **rpc_kwargs)
            return {entry.path: DataType(entry.metadata.data_type) for entry in entries}
        return {}


class SubscriberManager:
    def __init__(self, client: VSSClient):
        self.client = client
        self.subscribers = {}

    async def add_subscriber(
        self,
        subscribe_response_stream: AsyncIterator[List[EntryUpdate]],
        callback: Callable[[Iterable[EntryUpdate]], None],
    ) -> uuid.UUID:
        # We expect the first SubscribeResponse to be immediately available and to only hold a status
        await subscribe_response_stream.__aiter__().__anext__()  # pylint: disable=unnecessary-dunder-call

        sub_id = uuid.uuid4()
        new_sub_task = asyncio.create_task(
            self._subscriber_loop(subscribe_response_stream, callback)
        )
        self.subscribers[sub_id] = new_sub_task
        return sub_id

    async def remove_subscriber(self, subscription_id: uuid.UUID):
        try:
            subscriber_task = self.subscribers.pop(subscription_id)
        except KeyError as exc:
            raise ValueError(
                f"Could not find subscription {str(subscription_id)}"
            ) from exc
        subscriber_task.cancel()
        try:
            await subscriber_task
        except asyncio.CancelledError:
            pass

    async def _subscriber_loop(
        self,
        subscribe_response_stream: AsyncIterator[List[EntryUpdate]],
        callback: Callable[[Iterable[EntryUpdate]], None],
    ):
        async for updates in subscribe_response_stream:
            callback(updates)



#! /usr/bin/env python
########################################################################
# Copyright (c) 2020 Robert Bosch GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
########################################################################

import argparse
import functools
import json
import logging.config
import logging
import os
import pathlib
import sys
import threading
import time

from pygments import highlight
from pygments import lexers
from pygments import formatters
from cmd2 import Cmd
from cmd2 import CompletionItem
from cmd2 import with_argparser
from cmd2 import with_category
from cmd2 import constants
from cmd2.utils import basic_complete
from urllib.parse import urlparse

from kuksa_client import KuksaClientThread
from kuksa_client import _metadata
from kuksa_client.kuksa_logger import KuksaLogger

scriptDir = os.path.dirname(os.path.realpath(__file__))

DEFAULT_KUKSA_ADDRESS = os.environ.get("KUKSA_ADDRESS", "grpc://127.0.0.1:55555")
DEFAULT_TOKEN_OR_TOKENFILE = os.environ.get("TOKEN_OR_TOKENFILE", None)
DEFAULT_CACERTIFICATE = os.environ.get("CACERTIFICATE", None)
DEFAULT_TLS_SERVER_NAME = os.environ.get("TLS_SERVER_NAME", None)


logger = logging.getLogger(__name__)


def assignment_statement(arg):
    path, value = arg.split("=", maxsplit=1)
    return (path, value)


def display_completions(completions, delimiter):
    # Index of what prefix to remove from displayed items
    # I.e. "Vehicle." should be removed if the common prefix is "Vehicle.Ve".
    prefix_idx = os.path.commonprefix(completions).rfind(delimiter) + 1
    matches = []
    for path in completions:
        path = path[prefix_idx:]
        # Only display completions up to (and including) the next delimiter.
        next_dot = path.find(delimiter)
        if next_dot != -1:
            path = path[: next_dot + 1]
        matches.append(path)
    return matches


def metadata_tree_to_dict(tree):
    def add_children(flattened_tree, path, value):
        if "children" in value:
            for child_path, value in value["children"].items():
                add_children(flattened_tree, f"{path}.{child_path}", value)
        else:
            flattened_tree[path] = value

    flattened_tree = {}

    for key, value in tree.items():
        add_children(flattened_tree, key, value)

    return flattened_tree


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods
class TestClient(Cmd):
    def refresh_metadata(self):
        if self.server.startswith("grpc"):
            entries = json.loads(self.getMetaData("**"))
            if "error" in entries:
                raise Exception("Wrong databroker version, please use a newer version")
            # Convert to dict with paths as key
            self.metadata = {entry["path"]: entry for entry in entries}
        else:
            entries = json.loads(self.getMetaData(""))
            if "metadata" in entries:
                # Convert to dict with paths as key
                self.metadata = metadata_tree_to_dict(entries["metadata"])

    def path_completer(self, text, line, begidx, endidx):
        if not self.connection_established():
            return None

        if len(self.pathCompletionItems) == 0:
            self.refresh_metadata()

        # Normalize the delimiter used
        delimiter = "."
        if "/" in text:
            delimiter = "/"
            text = text.replace(delimiter, ".")

        # Generate the list of all possible completions
        self.pathCompletionItems = []
        for path in self.metadata.keys():
            # Compare case insensitive
            if path.lower().startswith(text.lower()):
                if delimiter != ".":
                    path = path.replace(".", delimiter)
                self.pathCompletionItems.append(path)

        # Generate the list of completions to display
        self.display_matches = display_completions(self.pathCompletionItems, delimiter)

        # Return all completions
        return self.pathCompletionItems

    def subscribeCallback(self, logPath, resp):
        if logPath is None:
            with self.terminal_lock:
                self.async_alert(
                    highlight(
                        json.dumps(json.loads(resp), indent=2),
                        lexers.JsonLexer(),
                        formatters.TerminalFormatter(),
                    )
                )
        else:
            with logPath.open("a", encoding="utf-8") as logFile:
                logFile.write(resp + "\n")

    def subscriptionIdCompleter(self, text, line, begidx, endidx):
        self.pathCompletionItems = []
        for sub_id in self.subscribeIds:
            self.pathCompletionItems.append(CompletionItem(sub_id))
        return basic_complete(text, line, begidx, endidx, self.pathCompletionItems)

    COMM_SETUP_COMMANDS = "Communication Set-up Commands"
    VSS_COMMANDS = "Kuksa Interaction Commands (Supported by both KUKSA Databroker and KUKSA Server)"
    VSS_COMMANDS_SERVER = "Kuksa Interaction Commands (Only supported by KUKSA Server)"
    INFO_COMMANDS = "Info Commands"

    ap_connect = argparse.ArgumentParser()
    ap_connect.add_argument(
        "server",
        help=f"VSS server to connect to. Format: protocol://host[:port]. \
        Supported protocols: [grpc, grpcs, ws, wss]. Example: {DEFAULT_KUKSA_ADDRESS}",
    )

    ap_disconnect = argparse.ArgumentParser()
    ap_authorize = argparse.ArgumentParser()
    tokenfile_completer_method = functools.partial(
        Cmd.path_complete,
        path_filter=lambda path: (os.path.isdir(path) or path.endswith(".token")),
    )
    ap_authorize.add_argument(
        "token_or_tokenfile",
        help="JWT(or the file storing the token) for authorizing the client.",
        completer_method=tokenfile_completer_method,
    )

    ap_setValue = argparse.ArgumentParser()
    ap_setValue.add_argument(
        "Path", help="Path to be set", completer_method=path_completer
    )
    ap_setValue.add_argument("Value", nargs="+", help="Value to be set")
    ap_setValue.add_argument(
        "-a", "--attribute", help="Attribute to be set", default="value"
    )

    ap_setValues = argparse.ArgumentParser()
    ap_setValues.add_argument(
        "Path=Value",
        help="Path and new value this path is to be set with",
        nargs="+",
        type=assignment_statement,
    )
    ap_setValues.add_argument(
        "-a", "--attribute", help="Attribute to be set", default="value"
    )

    ap_getValue = argparse.ArgumentParser()
    ap_getValue.add_argument(
        "Path", help="Path to be read", completer_method=path_completer
    )
    ap_getValue.add_argument(
        "-a", "--attribute", help="Attribute to be get", default="value"
    )

    ap_getValues = argparse.ArgumentParser()
    ap_getValues.add_argument(
        "Path",
        help="Path whose value is to be read",
        nargs="+",
        completer_method=path_completer,
    )
    ap_getValues.add_argument(
        "-a", "--attribute", help="Attribute to be get", default="value"
    )

    ap_setTargetValue = argparse.ArgumentParser()
    ap_setTargetValue.add_argument(
        "Path",
        help="Path whose target value to be set",
        completer_method=path_completer,
    )
    ap_setTargetValue.add_argument("Value", help="Value to be set")

    ap_setTargetValues = argparse.ArgumentParser()
    ap_setTargetValues.add_argument(
        "Path=Value",
        help="Path and new target value this path is to be set with",
        nargs="+",
        type=assignment_statement,
    )

    ap_getTargetValue = argparse.ArgumentParser()
    ap_getTargetValue.add_argument(
        "Path",
        help="Path whose target value is to be read",
        completer_method=path_completer,
    )

    ap_getTargetValues = argparse.ArgumentParser()
    ap_getTargetValues.add_argument(
        "Path",
        help="Path whose target value is to be read",
        nargs="+",
        completer_method=path_completer,
    )

    ap_subscribe = argparse.ArgumentParser()
    ap_subscribe.add_argument(
        "Path", help="Path to subscribe to", completer_method=path_completer
    )
    ap_subscribe.add_argument(
        "-a", "--attribute", help="Attribute to subscribe to", default="value"
    )

    ap_subscribe.add_argument(
        "-f",
        "--output-to-file",
        help="Redirect the subscription output to file",
        action="store_true",
    )

    ap_subscribeMultiple = argparse.ArgumentParser()
    ap_subscribeMultiple.add_argument(
        "Path", help="Path to subscribe to", nargs="+", completer_method=path_completer
    )
    ap_subscribeMultiple.add_argument(
        "-a", "--attribute", help="Attribute to subscribe to", default="value"
    )
    ap_subscribeMultiple.add_argument(
        "-f",
        "--output-to-file",
        help="Redirect the subscription output to file",
        action="store_true",
    )

    ap_unsubscribe = argparse.ArgumentParser()
    ap_unsubscribe.add_argument(
        "SubscribeId",
        help="Corresponding subscription Id",
        completer_method=subscriptionIdCompleter,
    )

    ap_getMetaData = argparse.ArgumentParser()
    ap_getMetaData.add_argument(
        "Path",
        help="Path whose metadata is to be read",
        completer_method=path_completer,
    )
    ap_updateMetaData = argparse.ArgumentParser()
    ap_updateMetaData.add_argument(
        "Path", help="Path whose MetaData is to update", completer_method=path_completer
    )
    ap_updateMetaData.add_argument(
        "Json",
        help="MetaData to update. Note, only attributes can be update, if update children or the whole vss tree, use"
        " `updateVSSTree` instead.",
    )

    ap_updateVSSTree = argparse.ArgumentParser()
    jsonfile_completer_method = functools.partial(
        Cmd.path_complete,
        path_filter=lambda path: (os.path.isdir(path) or path.endswith(".json")),
    )
    ap_updateVSSTree.add_argument(
        "Json",
        help="Json tree to update VSS",
        completer_method=jsonfile_completer_method,
    )

    # Constructor, request names after protocol to avoid errors
    def __init__(
        self,
        server=None,
        token_or_tokenfile=None,
        cacertificate=None,
        tls_server_name=None,
    ):
        shortcuts = constants.DEFAULT_SHORTCUTS
        shortcuts.update({"exit": "quit"})
        super().__init__(
            persistent_history_file=".vssclient_history",
            persistent_history_length=100,
            shortcuts=shortcuts,
            allow_cli_args=False,
        )

        self.prompt = "Test Client> "
        self.max_completion_items = 20
        self.server = server or DEFAULT_KUKSA_ADDRESS

        self.metadata = {}
        self.pathCompletionItems = []
        self.subscribeIds = set()
        self.commThread = None
        self.token_or_tokenfile = token_or_tokenfile
        self.cacertificate = cacertificate
        self.tls_server_name = tls_server_name

        with (pathlib.Path(scriptDir) / "logo").open("r", encoding="utf-8") as f:
            logo = f.read()
            print(logo.replace("%ver%", str(_metadata.__version__)))

        print()
        self.connect()

    @with_category(COMM_SETUP_COMMANDS)
    @with_argparser(ap_authorize)
    def do_authorize(self, args):
        """Authorize the client to interact with the server"""
        if args.token_or_tokenfile is not None:
            self.token_or_tokenfile = args.token_or_tokenfile
        if self.connection_established():
            resp = self.commThread.authorize(self.token_or_tokenfile)
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_setValue)
    def do_setValue(self, args):
        """Set the value of a path"""
        if self.connection_established():
            # If there is a blank before a single/double quote on the kuksa-client cli then
            # the argparser shell will remove it, there is nothing we can do to it
            # This gives off behavior for examples like:
            # setValue Vehicle.OBD.DTCList [ "dtc1, dtc2", ddd]
            # which will be treated as input of 3 elements
            # The recommended approach is to have quotes (of a different type) around the whole value
            # if your strings includes quotes, commas or other items
            # setValue Vehicle.OBD.DTCList '[ "dtc1, dtc2", ddd]'
            # or
            # setValue Vehicle.OBD.DTCList "[ 'dtc1, dtc2', ddd]"
            # If you really need to include a quote in the values use backslash and use the quote type
            # you want as inner value:
            # setValue Vehicle.OBD.DTCList "[ 'dtc1, \'dtc2', ddd]"
            # Will result in two elements in the array; "dtc1, 'dtc2" and "ddd"
            value = str(" ".join(args.Value))
            resp = self.commThread.setValue(args.Path, value, args.attribute)
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_setValues)
    def do_setValues(self, args):
        """Set the value of given paths"""
        if self.connection_established():
            resp = self.commThread.setValues(
                dict(getattr(args, "Path=Value")), args.attribute
            )
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_setTargetValue)
    def do_setTargetValue(self, args):
        """Set the target value of a path"""
        if self.connection_established():
            resp = self.commThread.setValue(args.Path, args.Value, "targetValue")
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_setTargetValues)
    def do_setTargetValues(self, args):
        """Set the target value of given paths"""
        if self.connection_established():
            resp = self.commThread.setValues(
                dict(getattr(args, "Path=Value")), "targetValue"
            )
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_getValue)
    def do_getValue(self, args):
        """Get the value of a path"""
        if self.connection_established():
            resp = self.commThread.getValue(args.Path, args.attribute)
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_getValues)
    def do_getValues(self, args):
        """Get the value of given paths"""
        if self.connection_established():
            resp = self.commThread.getValues(args.Path, args.attribute)
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_getTargetValue)
    def do_getTargetValue(self, args):
        """Get the value of a path"""
        if self.connection_established():
            resp = self.commThread.getValue(args.Path, "targetValue")
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_getTargetValues)
    def do_getTargetValues(self, args):
        """Get the value of given paths"""
        if self.connection_established():
            resp = self.commThread.getValues(args.Path, "targetValue")
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_subscribe)
    def do_subscribe(self, args):
        """Subscribe the value of a path"""
        if self.connection_established():
            if args.output_to_file:
                logPath = (
                    pathlib.Path.cwd()
                    / f"log_{args.Path.replace('/', '.')}_{args.attribute}_{str(time.time())}"
                )
                callback = functools.partial(self.subscribeCallback, logPath)
            else:
                callback = functools.partial(self.subscribeCallback, None)

            resp = self.commThread.subscribe(args.Path, callback, args.attribute)
            resJson = json.loads(resp)
            if "subscriptionId" in resJson:
                self.subscribeIds.add(resJson["subscriptionId"])
                if args.output_to_file:
                    logPath.touch()
                    print(f"Subscription log available at {logPath}")
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_subscribeMultiple)
    def do_subscribeMultiple(self, args):
        """Subscribe to updates of given paths"""
        if self.connection_established():
            if args.output_to_file:
                logPath = (
                    pathlib.Path.cwd()
                    / f"subscribeMultiple_{args.attribute}_{str(time.time())}.log"
                )
                callback = functools.partial(self.subscribeCallback, logPath)
            else:
                callback = functools.partial(self.subscribeCallback, None)
            resp = self.commThread.subscribeMultiple(
                args.Path, callback, args.attribute
            )
            resJson = json.loads(resp)
            if "subscriptionId" in resJson:
                self.subscribeIds.add(resJson["subscriptionId"])
                if args.output_to_file:
                    logPath.touch()
                    print(f"Subscription log available at {logPath}")
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_unsubscribe)
    def do_unsubscribe(self, args):
        """Unsubscribe an existing subscription"""
        if self.connection_established():
            resp = self.commThread.unsubscribe(args.SubscribeId)
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
            self.subscribeIds.discard(args.SubscribeId)
            self.pathCompletionItems = []

    def stop(self):
        if self.commThread is not None:
            self.commThread.stop()
            self.commThread.join()

    def getMetaData(self, path):
        """Get MetaData of the path"""
        if self.connection_established():
            return self.commThread.getMetaData(path)
        return "{}"

    @with_category(VSS_COMMANDS_SERVER)
    @with_argparser(ap_updateVSSTree)
    def do_updateVSSTree(self, args):
        """Update VSS Tree Entry"""
        if self.connection_established():
            resp = self.commThread.updateVSSTree(args.Json)
            if resp is not None:
                print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_updateMetaData)
    def do_updateMetaData(self, args):
        """Update MetaData of a given path"""
        if self.connection_established():
            resp = self.commThread.updateMetaData(args.Path, args.Json)
            print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))

    @with_category(VSS_COMMANDS)
    @with_argparser(ap_getMetaData)
    def do_getMetaData(self, args):
        """Get MetaData of the path"""
        resp = self.getMetaData(args.Path)
        print(highlight(resp, lexers.JsonLexer(), formatters.TerminalFormatter()))
        self.pathCompletionItems = []

    @with_category(COMM_SETUP_COMMANDS)
    @with_argparser(ap_disconnect)
    def do_disconnect(self, _args):
        """Disconnect from the VISS/gRPC Server"""
        if hasattr(self, "commThread"):
            if self.commThread is not None:
                self.commThread.stop()
                self.commThread = None

    def connection_established(self) -> bool:
        """
        Check if thread has established a connection to the broker/server.
        Note that this method does not indicate the current state of the connection,
        This method may return True even if the broker/server currently is not reachable.
        """
        if self.commThread is None or not self.commThread.connection_established():
            self.connect()
        return self.commThread.connection_established()

    def connect(self):
        """Connect to the VISS/gRPC Server"""
        if hasattr(self, "commThread"):
            if self.commThread is not None:
                self.commThread.stop()
                self.commThread = None

        # Check we have a valid server URI
        srv = urlparse(self.server)
        config = {"port": 55555, "insecure": True}

        if srv.scheme in ["grpc", "grpcs"]:
            config["protocol"] = "grpc"
        elif srv.scheme in ["ws", "wss"]:
            config["protocol"] = "ws"
            config["port"] = 8090
        else:
            print(f"Invalid server URI. Unsupported protocol: {srv.scheme} ")
            return

        if srv.port is not None:
            config["port"] = srv.port

        if srv.scheme in ["grpcs", "wss"]:
            if self.cacertificate is None:
                print("TLS cannot be used as no CA Certificate specifed!")
            else:
                config["insecure"] = False

        if srv.hostname is None:
            print("No hostname or IP given")
            return

        config["ip"] = srv.hostname

        # Explain were we are connecting to:
        print(
            f"Connecting to VSS server at {config['ip'] } port {config['port'] } \
using {'KUKSA GRPC' if config['protocol'] == 'grpc' else 'VISS' } protocol."
        )
        print(f"TLS will {'not be' if config['insecure'] else 'be'} used.")

        # Configs should only be added if they actually have a value
        if self.token_or_tokenfile is not None:
            config["token_or_tokenfile"] = self.token_or_tokenfile
        if self.cacertificate is not None:
            config["cacertificate"] = self.cacertificate
        if self.tls_server_name is not None:
            config["tls_server_name"] = self.tls_server_name

        self.commThread = KuksaClientThread(config)
        self.commThread.start()

        waitForConnection = threading.Condition()
        waitForConnection.acquire()
        waitForConnection.wait_for(self.commThread.connection_established, timeout=1)
        waitForConnection.release()

        if self.commThread.connection_established():
            pass
        else:
            print(
                "Error: Websocket could not be connected or the gRPC channel could not be created."
            )
            self.commThread.stop()
            self.commThread = None

    @with_category(COMM_SETUP_COMMANDS)
    @with_argparser(ap_connect)
    def do_connect(self, args):
        """Connect to a VSS server"""
        self.server = args.server
        self.connect()

    @with_category(INFO_COMMANDS)
    def do_info(self, _args):
        """Show summary info of the client"""
        print("kuksa-client version " + _metadata.__version__)
        print("Uri: " + _metadata.__uri__)
        print("Author: " + _metadata.__author__)
        print("Copyright: " + _metadata.__copyright__)

    @with_category(INFO_COMMANDS)
    def do_version(self, _args):
        """Show version of the client"""
        print(_metadata.__version__)


# pylint: enable=too-many-public-methods
# pylint: enable=too-many-instance-attributes

# Main Function

def main():

    kuksa_logger = KuksaLogger()
    kuksa_logger.init_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "server",
        nargs="?",
        help=f"VSS server to connect to. Format: protocol://host[:port]. \
        Supported protocols: [grpc, grpcs, ws, wss]. Example: {DEFAULT_KUKSA_ADDRESS}",
        default=DEFAULT_KUKSA_ADDRESS,
    )
    parser.add_argument(
        "--token_or_tokenfile",
        default=DEFAULT_TOKEN_OR_TOKENFILE,
        help="JWT token or path to a JWT token file (.token)",
    )

    # Add TLS arguments
    parser.add_argument(
        "--cacertificate",
        default=DEFAULT_CACERTIFICATE,
        help="Client root cert file (.pem). \
        Needed for TLS enabled transports (grpcs, wss)",
    )
    # Observations for Python
    # Connecting to "localhost" works well, subjectAltName seems to suffice
    # Connecting to "127.0.0.1" does not work unless server-name specified
    # For KUKSA.val example certs default name is "Server"
    parser.add_argument(
        "--tls-server-name",
        default=DEFAULT_TLS_SERVER_NAME,
        help="CA name of server, needed in some cases where subjectAltName does not suffice",
    )

    args = parser.parse_args()

    clientApp = TestClient(
        args.server,
        token_or_tokenfile=args.token_or_tokenfile,
        cacertificate=args.cacertificate,
        tls_server_name=args.tls_server_name,
    )
    try:
        # We exit the loop when the user types "quit" or hits Ctrl-D.
        clientApp.cmdloop()
    finally:
        clientApp.stop()


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: kuksa/val/v1/val.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'kuksa/val/v1/val.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from kuksa.val.v1 import types_pb2 as kuksa_dot_val_dot_v1_dot_types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16kuksa/val/v1/val.proto\x12\x0ckuksa.val.v1\x1a\x18kuksa/val/v1/types.proto\"c\n\x0c\x45ntryRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12 \n\x04view\x18\x02 \x01(\x0e\x32\x12.kuksa.val.v1.View\x12#\n\x06\x66ields\x18\x03 \x03(\x0e\x32\x13.kuksa.val.v1.Field\"9\n\nGetRequest\x12+\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x1a.kuksa.val.v1.EntryRequest\"\x89\x01\n\x0bGetResponse\x12(\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x17.kuksa.val.v1.DataEntry\x12,\n\x06\x65rrors\x18\x02 \x03(\x0b\x32\x1c.kuksa.val.v1.DataEntryError\x12\"\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x13.kuksa.val.v1.Error\"Z\n\x0b\x45ntryUpdate\x12&\n\x05\x65ntry\x18\x01 \x01(\x0b\x32\x17.kuksa.val.v1.DataEntry\x12#\n\x06\x66ields\x18\x02 \x03(\x0e\x32\x13.kuksa.val.v1.Field\"8\n\nSetRequest\x12*\n\x07updates\x18\x01 \x03(\x0b\x32\x19.kuksa.val.v1.EntryUpdate\"_\n\x0bSetResponse\x12\"\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x13.kuksa.val.v1.Error\x12,\n\x06\x65rrors\x18\x02 \x03(\x0b\x32\x1c.kuksa.val.v1.DataEntryError\"C\n\x15StreamedUpdateRequest\x12*\n\x07updates\x18\x01 \x03(\x0b\x32\x19.kuksa.val.v1.EntryUpdate\"j\n\x16StreamedUpdateResponse\x12\"\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x13.kuksa.val.v1.Error\x12,\n\x06\x65rrors\x18\x02 \x03(\x0b\x32\x1c.kuksa.val.v1.DataEntryError\"e\n\x0eSubscribeEntry\x12\x0c\n\x04path\x18\x01 \x01(\t\x12 \n\x04view\x18\x02 \x01(\x0e\x32\x12.kuksa.val.v1.View\x12#\n\x06\x66ields\x18\x03 \x03(\x0e\x32\x13.kuksa.val.v1.Field\"A\n\x10SubscribeRequest\x12-\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x1c.kuksa.val.v1.SubscribeEntry\"?\n\x11SubscribeResponse\x12*\n\x07updates\x18\x01 \x03(\x0b\x32\x19.kuksa.val.v1.EntryUpdate\"\x16\n\x14GetServerInfoRequest\"6\n\x15GetServerInfoResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t2\x88\x03\n\x03VAL\x12:\n\x03Get\x12\x18.kuksa.val.v1.GetRequest\x1a\x19.kuksa.val.v1.GetResponse\x12:\n\x03Set\x12\x18.kuksa.val.v1.SetRequest\x1a\x19.kuksa.val.v1.SetResponse\x12_\n\x0eStreamedUpdate\x12#.kuksa.val.v1.StreamedUpdateRequest\x1a$.kuksa.val.v1.StreamedUpdateResponse(\x01\x30\x01\x12N\n\tSubscribe\x12\x1e.kuksa.val.v1.SubscribeRequest\x1a\x1f.kuksa.val.v1.SubscribeResponse0\x01\x12X\n\rGetServerInfo\x12\".kuksa.val.v1.GetServerInfoRequest\x1a#.kuksa.val.v1.GetServerInfoResponseB\x0eZ\x0ckuksa/val/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kuksa.val.v1.val_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\014kuksa/val/v1'
  _globals['_ENTRYREQUEST']._serialized_start=66
  _globals['_ENTRYREQUEST']._serialized_end=165
  _globals['_GETREQUEST']._serialized_start=167
  _globals['_GETREQUEST']._serialized_end=224
  _globals['_GETRESPONSE']._serialized_start=227
  _globals['_GETRESPONSE']._serialized_end=364
  _globals['_ENTRYUPDATE']._serialized_start=366
  _globals['_ENTRYUPDATE']._serialized_end=456
  _globals['_SETREQUEST']._serialized_start=458
  _globals['_SETREQUEST']._serialized_end=514
  _globals['_SETRESPONSE']._serialized_start=516
  _globals['_SETRESPONSE']._serialized_end=611
  _globals['_STREAMEDUPDATEREQUEST']._serialized_start=613
  _globals['_STREAMEDUPDATEREQUEST']._serialized_end=680
  _globals['_STREAMEDUPDATERESPONSE']._serialized_start=682
  _globals['_STREAMEDUPDATERESPONSE']._serialized_end=788
  _globals['_SUBSCRIBEENTRY']._serialized_start=790
  _globals['_SUBSCRIBEENTRY']._serialized_end=891
  _globals['_SUBSCRIBEREQUEST']._serialized_start=893
  _globals['_SUBSCRIBEREQUEST']._serialized_end=958
  _globals['_SUBSCRIBERESPONSE']._serialized_start=960
  _globals['_SUBSCRIBERESPONSE']._serialized_end=1023
  _globals['_GETSERVERINFOREQUEST']._serialized_start=1025
  _globals['_GETSERVERINFOREQUEST']._serialized_end=1047
  _globals['_GETSERVERINFORESPONSE']._serialized_start=1049
  _globals['_GETSERVERINFORESPONSE']._serialized_end=1103
  _globals['_VAL']._serialized_start=1106
  _globals['_VAL']._serialized_end=1498
# @@protoc_insertion_point(module_scope)



########################################################################
# Copyright (c) 2022 Robert Bosch GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
########################################################################


import asyncio
import json
import pathlib
import queue
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Optional
import uuid
import os
import logging

from kuksa_client import cli_backend
import kuksa_client.grpc
import kuksa_client.grpc.aio
from kuksa_client.grpc import EntryUpdate
from kuksa.val.v1 import types_pb2


logger = logging.getLogger(__name__)


def callback_wrapper(callback: Callable[[str], None]) -> Callable[[Iterable[EntryUpdate]], None]:
    def wrapper(updates: Iterable[EntryUpdate]) -> None:
        try:
            callback(json.dumps([update.to_dict() for update in updates], cls=DatabrokerEncoder))
        except Exception as e:
            logger.error("Callback could not be executed", e)
    return wrapper


class DatabrokerEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (types_pb2.StringArray, types_pb2.BoolArray, types_pb2.Uint32Array, types_pb2.Uint64Array,
                            types_pb2.FloatArray, types_pb2.Int32Array, types_pb2.Int64Array, types_pb2.DoubleArray)):
            string_values = []
            for value in obj.values:
                value = str(value)
                string_values.append(value)
            return {'values': string_values}
        return super().default(obj)


class Backend(cli_backend.Backend):
    def __init__(self, config):
        super().__init__(config)
        if self.cacertificate is not None:
            self.cacertificate = pathlib.Path(self.cacertificate)
        if self.token_or_tokenfile is not None:
            if os.path.isfile(self.token_or_tokenfile):
                self.token_or_tokenfile = pathlib.Path(self.token_or_tokenfile)
                self.token = self.token_or_tokenfile.expanduser(
                ).read_text(encoding='utf-8').rstrip('\n')
            else:
                self.token = str(self.token_or_tokenfile)
        else:
            self.token = ""
        self.grpc_connection_established = False

        self.sendMsgQueue = queue.Queue()
        self.run = False

        self.AttrDict = {
            "value": (
                kuksa_client.grpc.Field.VALUE,
                kuksa_client.grpc.View.CURRENT_VALUE,
            ),
            "targetValue": (
                kuksa_client.grpc.Field.ACTUATOR_TARGET,
                kuksa_client.grpc.View.TARGET_VALUE,
            ),
            "metadata": (
                kuksa_client.grpc.Field.METADATA,
                kuksa_client.grpc.View.METADATA,
            ),
        }

    def connection_established(self) -> bool:
        """
        Function to check connection status
        """
        return self.grpc_connection_established

    # Function to stop the communication
    def stop(self):
        self.disconnect()
        self.run = False
        logger.info("gRPC channel disconnected.")

    # Function to implement fetching of metadata
    def getMetaData(self, path: str, timeout=5):
        return self.getValue(path, "metadata", timeout)

    def updateMetaData(self, path: str, jsonStr, timeout=5):
        return self.setValue(path, jsonStr, "metadata", timeout)

    # Function to implement get
    def getValue(self, path: str, attribute="value", timeout=5):
        return self.getValues((path,), attribute, timeout)

    def getValues(self, paths: Iterable[str], attribute="value", timeout=5):
        if attribute in self.AttrDict:
            field, view = self.AttrDict[attribute]
            entries = [
                kuksa_client.grpc.EntryRequest(path=path, view=view, fields=(field,))
                for path in paths
            ]
            requestArgs = {'entries': entries}
            return self._sendReceiveMsg(("get", requestArgs), timeout)

        return json.dumps({"error": "Invalid Attribute"})

    # Function to implement set
    def setValue(self, path: str, value, attribute="value", timeout=5):
        return self.setValues({path: value}, attribute, timeout)

    def setValues(self, updates: Dict[str, Any], attribute="value", timeout=5):
        if attribute in self.AttrDict:
            field, _ = self.AttrDict[attribute]
            entry_updates = []
            try_v2 = False
            for path, value in updates.items():

                if field is kuksa_client.grpc.Field.VALUE:
                    entry = kuksa_client.grpc.DataEntry(
                        path=path,
                        value=kuksa_client.grpc.Datapoint(value=value),
                    )
                    try_v2 = True
                elif field is kuksa_client.grpc.Field.ACTUATOR_TARGET:
                    entry = kuksa_client.grpc.DataEntry(
                        path=path,
                        actuator_target=kuksa_client.grpc.Datapoint(value=value),
                    )
                elif field is kuksa_client.grpc.Field.METADATA:
                    try:
                        metadata_dict = json.loads(value)
                    except json.JSONDecodeError:
                        return json.dumps({"error": "Metadata value needs to be a valid JSON object"})
                    entry = kuksa_client.grpc.DataEntry(
                        path=path,
                        metadata=kuksa_client.grpc.Metadata.from_dict(metadata_dict),
                    )
                entry_updates.append(
                    kuksa_client.grpc.EntryUpdate(entry=entry, fields=(field,))
                )
            requestArgs = {"updates": entry_updates, "try_v2": try_v2}
            return self._sendReceiveMsg(("set", requestArgs), timeout)
        return json.dumps({"error": "Invalid Attribute"})

    # Function for authorization
    def authorize(self, token_or_tokenfile: Optional[str] = None, timeout=5):
        if token_or_tokenfile is None:
            token_or_tokenfile = self.token_or_tokenfile
        if os.path.isfile(token_or_tokenfile):
            token_or_tokenfile_path = pathlib.Path(token_or_tokenfile)
            token = token_or_tokenfile_path.expanduser().read_text(encoding='utf-8').rstrip('\n')
        else:
            token = token_or_tokenfile
        requestArgs = {
            'token': token
        }

        return self._sendReceiveMsg(("authorize", requestArgs), timeout)

    # Subscribe value changes of to a given path.
    # The given callback function will be called then, if the given path is updated.
    def subscribe(self, path: str, callback, attribute="value", timeout=5):
        return self.subscribeMultiple((path,), callback, attribute, timeout)

    def subscribeMultiple(self, paths: Iterable[str], callback, attribute="value", timeout=5):
        if attribute in self.AttrDict:
            field, view = self.AttrDict[attribute]
            entries = [
                kuksa_client.grpc.SubscribeEntry(path=path, view=view, fields=(field,))
                for path in paths
            ]
            requestArgs = {
                "entries": entries,
                "try_v2": True,
                "callback": callback_wrapper(callback),
            }
            return self._sendReceiveMsg(("subscribe", requestArgs), timeout)

        return json.dumps({"error": "Invalid Attribute"})

    # Unsubscribe value changes of to a given path.
    # The subscription id from the response of the corresponding subscription request will be required
    def unsubscribe(self, sub_id: str, timeout=5):
        try:
            sub_uuid = uuid.UUID(sub_id)
        except ValueError as exc:
            return json.dumps({"error": str(exc)})
        requestArgs = {'subscription_id': sub_uuid}
        return self._sendReceiveMsg(("unsubscribe", requestArgs), timeout)

    def connect(self, timeout=5):
        requestArgs = {}
        return self._sendReceiveMsg(("connect", requestArgs), timeout)

    def disconnect(self, timeout=5):
        requestArgs = {}
        return self._sendReceiveMsg(("disconnect", requestArgs), timeout)

    def _sendReceiveMsg(self, req, timeout):
        (call, requestArgs) = req
        recvQueue = queue.Queue(maxsize=1)
        self.sendMsgQueue.put((call, requestArgs, recvQueue))
        try:
            resp, error = recvQueue.get(timeout=timeout)
            if error:
                respJson = json.dumps(error, indent=4)
            elif resp:
                respJson = json.dumps(resp, indent=4, cls=DatabrokerEncoder)
            else:
                respJson = "OK"
        except queue.Empty:
            respJson = json.dumps({"error": "Timeout"})
        return respJson

    # Async function to handle the gRPC calls
    async def _grpcHandler(self, vss_client: kuksa_client.grpc.aio.VSSClient):
        self.run = True
        subscriber_manager = kuksa_client.grpc.aio.SubscriberManager(vss_client)
        self.grpc_connection_established = True
        while self.run:
            try:
                (call, requestArgs, responseQueue) = self.sendMsgQueue.get_nowait()
            except queue.Empty:
                await asyncio.sleep(0.01)
                continue
            try:
                if call == "get":
                    resp = await vss_client.get(**requestArgs)
                    if resp is not None:
                        resp = [entry.to_dict() for entry in resp]
                        resp = resp[0] if len(resp) == 1 else resp
                elif call == "set":
                    resp = await vss_client.set(**requestArgs)
                elif call == "authorize":
                    resp = await vss_client.authorize(str(requestArgs["token"]))
                elif call == "subscribe":
                    callback = requestArgs.pop('callback')
                    subscriber_response_stream = vss_client.subscribe(
                        **requestArgs)
                    resp = await subscriber_manager.add_subscriber(subscriber_response_stream, callback)
                    resp = {"subscriptionId": str(resp)}
                elif call == "unsubscribe":
                    resp = await subscriber_manager.remove_subscriber(**requestArgs)
                elif call == "connect":
                    resp = await vss_client.connect()
                elif call == "disconnect":
                    resp = await vss_client.disconnect()
                else:
                    raise Exception("Not Implemented.")

                responseQueue.put((resp, None))
            except kuksa_client.grpc.VSSClientError as exc:
                responseQueue.put((None, exc.to_dict()))
            except ValueError:
                responseQueue.put(
                    (None, {"error": "ValueError in casting the value."}))

        self.grpc_connection_established = False

    # Update VSS Tree Entry
    def updateVSSTree(self, jsonStr, timeout=5):
        logger.warning("Command not supported by KUKSA Databroker or KUKSA gRPC!")

    # Main loop for handling gRPC communication
    async def mainLoop(self):
        if self.insecure:

            async with kuksa_client.grpc.aio.VSSClient(
                self.serverIP, self.serverPort, token=self.token
            ) as vss_client:
                logger.info("gRPC channel connected.")
                await self._grpcHandler(vss_client)
        else:
            async with kuksa_client.grpc.aio.VSSClient(
                self.serverIP,
                self.serverPort,
                root_certificates=self.cacertificate,
                tls_server_name=self.tls_server_name,
                token=self.token,
            ) as vss_client:
                logger.info("Secure gRPC channel connected.")
                await self._grpcHandler(vss_client)



#! /usr/bin/env python

########################################################################
# Copyright (c) 2020 Robert Bosch GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
########################################################################


import logging

__all__ = (
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
)

__title__ = "Unknown"
__summary__ = "Unknown"
__uri__ = "Unknown"
__version__ = "Unknown"
__author__ = "Unknown"
__email__ = "Unknown"
__license__ = "Unknown"
__copyright__ = "Copyright 2020 Robert Bosch GmbH"

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

try:

    metadata = importlib_metadata.metadata("kuksa_client")

    __title__ = metadata["name"]
    __summary__ = metadata["summary"]
    __uri__ = metadata["home-page"]
    __version__ = metadata["version"]
    __author__ = metadata["author"]
    __email__ = metadata["author-email"]
    __license__ = metadata["license"]

except importlib_metadata.PackageNotFoundError as e:
    logger = logging.getLogger(__name__)
    logger.info("skip configuring metadata", e)
